#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_length.py


import os
import pandas as pd
from collections import OrderedDict
import argparse


def merge_length_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the length distribution files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for length distribution files (e.g., '*length_distribution.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name.'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def process_log_files(length_file_list):
    '''
    @Message  : retrieve the length table.
    @Input    : list --> pattern for reads length files (e.g., '*length_distribution.txt')
    @Return   : 
                output --> output nested dict contain the reads length
    @Flow     : step1 --> retrieve the reads length
    '''

    # make the dataframe
    reads_length_df = pd.DataFrame()
    reads_length_df['Sample'] = []
    reads_length_df['Length'] = []
    reads_length_df['Plus_Count'] = []
    reads_length_df['Minus_Count'] = []
    reads_length_df['Plus_Ratio'] = []
    reads_length_df['Minus_Ratio'] = []

    # for each reads length file
    for length_file in length_file_list:
        file_prefix = os.path.basename(length_file).split('.')[0].replace('_length_distribution', '')

        # import the reads length file
        length_df = pd.read_csv(length_file, sep='\t', index_col=False)

        # rename the columns
        length_df.columns = ['Length', 'Plus_Count', 'Minus_Count']

        # calculate the ratio of reads length
        length_df['Plus_Ratio'] = length_df['Plus_Count'] / length_df['Plus_Count'].sum() * 100
        length_df['Minus_Ratio'] = length_df['Minus_Count'] / length_df['Minus_Count'].sum() * 100

        length_df['Sample'] = file_prefix

        # merge the reads length file
        reads_length_df = pd.concat([reads_length_df, length_df], axis=0)

    return reads_length_df


def output_table(reads_length_df, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the mapped reads
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    # keep the 2 decimal places
    reads_length_df[['Length', 'Plus_Count', 'Minus_Count']] = reads_length_df[['Length', 'Plus_Count', 'Minus_Count']].astype('int')
    reads_length_df[['Plus_Ratio', 'Minus_Ratio']] = reads_length_df[['Plus_Ratio', 'Minus_Ratio']].astype('float').round(4)

    reads_length_df.to_csv(output_file + '_length_distribution.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_length_args_parser()

    print('Step2: import the reads length file.', flush=True)
    reads_length_df = process_log_files(args.list)

    print('Step3: output the length distribution table.', flush=True)
    output_table(reads_length_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Ribo.py


import os
import sys
from collections import OrderedDict
from itertools import islice
from Bio import SeqIO
from math import ceil
import numpy as np
import pandas as pd
import polars as pl
import pysam


class Mrna(object):

    def __init__(self, record):
        self.chromosome = record[0]
        self.gene_id = record[1]
        # self.gene_name = record[2]
        self.transcript_id = record[2]
        self.start = record[3]
        self.end = record[4]
        self.utr5_length = int(record[5])
        self.cds_length = int(record[6])
        self.utr3_length = int(record[7])
        self.strand = record[8]
        self.rep_transcript = record[9]
        self.modified = record[10]
        self.bam = []
        self.rpf = []
        self.seq = None


class Ribo(object):

    def __init__(self, args):
        self.mrna_file = args.transcript
        self.mrna_seq = args.sequence
        self.mrna_dict = OrderedDict()
        self.longest = args.longest
        self.silence = args.silence

        self.bam = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]
        self.pysam_input = None

        self.psite = args.psite
        self.periodicity = args.periodicity
        self.min_length = args.min
        self.max_length = args.max
        self.offset = {}
        self.profile = None

        self.thread = args.thread

        self.total_rpf_df = None
        self.output = args.output

    def read_transcript(self):
        '''
        @Message  : import the gene sequence and message
        @Input    : mrna_seq --> the gene sequences
                    mrna_file --> the gene message
        @Return   : mrna_dict --> the gene message
        @Flow     : step1 --> import the mrna sequence generate from rpf_Reference script
                        contains: [5'-utr, cds, 3'-utr]
                    step2 --> import the gene message generate from rpf_Reference script
                        contains id same as the Mrna class
        '''

        # import the mrna sequences
        mrna_seq = SeqIO.parse(self.mrna_seq, "fasta")
        mrna_sequence = OrderedDict()
        for line in mrna_seq:
            # sys.stdout.writelines("import gene:  {gene}\n".format(gene=line.id))
            mrna_sequence[line.id] = line.seq

        # mrna_sequence = pysam.FastaFile(self.mrna_seq)

        # import the gene message
        with open(self.mrna_file, 'r') as trans_file_in:
            if self.longest:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    if int(record[6]) % 3 != 0 and not self.silence:
                        sys.stdout.write("{gene} CDS did not fit 3nt periodicity. \n".format(gene=record[2]))

                    if record[9] == "True":
                        now_mrna = Mrna(record)
                        # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                        now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                        try:
                            now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                            now_mrna.rpf = np.zeros(len(now_mrna.seq))
                            # now_mrna.rpf = [0] * len(now_mrna.seq)
                            self.mrna_dict[now_mrna.transcript_id] = now_mrna
                        except KeyError:
                            continue
                    else:
                        continue
            else:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    if int(record[6]) % 3 != 0:
                        sys.stdout.write("{gene} CDS did not fit 3nt periodicity. \n".format(gene=record[2]))

                    now_mrna = Mrna(record)
                    # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                    now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                    try:
                        now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                        now_mrna.rpf = np.zeros(len(now_mrna.seq))
                        self.mrna_dict[now_mrna.transcript_id] = now_mrna
                    except KeyError:
                        continue

    def read_offset(self):
        '''
        @Message  : import the offset and check the type of sequencing profile
        @Input    : psite --> the offset file
        @Return   : self.offset --> the offset dictionary
        @Flow     : step1 --> import the offset generate from the rpf_Offset script
                    step2 --> retrieve the psite form offset file
                    step3 --> check the type of sequencing profie from offset file
        '''

        offset = pd.read_csv(self.psite, sep='\t', header=0, names=None)
        offset.dropna(axis=0, inplace=True)
        offset["p_site"] = offset["p_site"].astype(int)
        offset["p_site"] = offset["p_site"] - 1

        raw_offset = offset[["length", "p_site"]].groupby(offset["length"])["p_site"].apply(list).T.to_dict()
        print("The import offset table: ", flush=True)
        print(raw_offset, flush=True)

        # filter the min_length, max_length and periodicity
        offset = offset[offset['periodicity'] >= self.periodicity]
        offset = offset[(offset["length"] >= self.min_length) & (offset["length"] <= self.max_length)]

        self.offset = offset[["length", "p_site"]].groupby(offset["length"])["p_site"].apply(list).T.to_dict()
        print("Filterd RPFs offset table: ", flush=True)
        print(self.offset, flush=True)

        # check the data type.
        if "third" in offset["ribo"].unique():
            self.profile = "trisome"
        elif "second" in offset["ribo"].unique():
            self.profile = "disome"
        elif "first" in offset["ribo"].unique():
            self.profile = "monosome"
        else:
            sys.stdout.writelines("Unknown type of sequence profile, please re-check the offset file!")
            sys.exit()

    def monosome_p_site(self):
        '''
        @Message  : retrieve the psite from each reads
        @Input    : pysam_input --> the mapping bam file interface
                    mrna_dict --> the gene message
        @Return   : output --> description
        @Flow     : step1 --> import the gene mapping bam file
                    step2 --> delete the reads not in the specified range
                    step3 --> retrieve the psite from each reads
        '''

        for line in self.pysam_input.fetch(until_eof=True):
            if line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()

                if (self.min_length <= read_length <= self.max_length) and (read_length in self.offset):
                    if self.offset[read_length][0] == -1:
                        p_site = line.get_reference_positions()
                        self.mrna_dict[line.reference_name].rpf[p_site] += 1

                    else:
                        p_site = map_start + self.offset[read_length][0]
                        self.mrna_dict[line.reference_name].rpf[p_site] += 1

                else:
                    pass

    def disome_p_site(self):
        """
        like the mono_p_site,
        parser the reads with function twice,
        """

        for line in self.pysam_input.fetch(until_eof=True):
            if line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()

                if (self.min_length <= read_length <= self.max_length) and (read_length in self.offset):

                    if self.offset[read_length][0] == -1:
                        p_site = map_start + np.arange(read_length)

                        for site in p_site:
                            try:
                                self.mrna_dict[line.reference_name].rpf[site] += 1
                            except KeyError:
                                continue
                            except IndexError:
                                continue
                    else:

                        try:
                            p_site = map_start + self.offset[read_length][0]
                            self.mrna_dict[line.reference_name].rpf[p_site] += 1
                            p_site = map_start + self.offset[read_length][1]
                            self.mrna_dict[line.reference_name].rpf[p_site] += 1
                        except KeyError:
                            # sys.stdout.writelines("skip reads {reads}!".format(reads=line))
                            continue
                        except IndexError:
                            continue
                else:
                    pass

    def trisome_p_site(self):
        """
        like the disome_p_site
        """

        for line in self.pysam_input.fetch(until_eof=True):
            if line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()

                if (self.min_length <= read_length <= self.max_length) and (read_length in self.offset):
                    if self.offset[read_length][0] == -1:
                        p_site = map_start + np.arange(read_length)

                        for site in p_site:
                            try:
                                self.mrna_dict[line.reference_name].rpf[site] += 1
                            except KeyError:
                                continue
                            except IndexError:
                                continue
                    else:

                        try:
                            p_site = map_start + self.offset[read_length][0]
                            self.mrna_dict[line.reference_name].rpf[p_site] += 1
                            p_site = map_start + self.offset[read_length][1]
                            self.mrna_dict[line.reference_name].rpf[p_site] += 1
                            p_site = map_start + self.offset[read_length][2]
                            self.mrna_dict[line.reference_name].rpf[p_site] += 1
                        except KeyError:
                            # sys.stdout.writelines("skip reads {reads}!".format(reads=line))
                            continue
                        except IndexError:
                            continue
                else:
                    pass

    @staticmethod
    def flt_star_results(args):
        '''
        @Message  : filter the bam file generated by STAR
        @Input    : bam_in_file --> the path of input bam file
                    bam_out_file --> the path of output bam file
                    gene_list --> the list of gene name
                    tag --> the number of unique reads
                    sample_format --> the format of bam file
        @Return   : length_dict --> the dict of length distribution
                    bam_seq_dict --> the dict of sequence profile
        @Flow     : step1: open the bam file
                    step2: read the bam file and store the information into a dict
                    step3: calculate the length distribution
                    step4: output the filtered bam file
        '''
        
        bam_in_file, mrna_dict, min_length, max_length, offset, sample_format = args
        
        with pysam.AlignmentFile(bam_in_file, sample_format) as bam_in:
            for gene in mrna_dict.keys():
                try:
                    gene_reads = bam_in.fetch(gene)
                except ValueError:
                    continue

                # filter the mapped reads or not
                for reads in gene_reads:
                    if reads.is_unmapped:
                        continue

                    if reads.reference_name in mrna_dict:
                        map_start, map_end = reads.get_blocks()[0]
                        read_length = reads.infer_read_length()

                        if (min_length <= read_length <= max_length) and (read_length in offset):
                            if offset[read_length][0] == -1:
                                p_site = map_start + np.arange(read_length)
                                for site in p_site:
                                    try:
                                        mrna_dict[reads.reference_name].rpf[site] += 1
                                    except KeyError:
                                        continue
                                    except IndexError:
                                        continue
                            else:

                                p_site = map_start + offset[read_length][0]
                                try:
                                    mrna_dict[reads.reference_name].rpf[p_site] += 1

                                except KeyError:
                                    continue
                                    # sys.stdout.writelines("skip reads {reads}!".format(reads=line))
                                except IndexError:
                                    continue
                                    # sys.stdout.writelines("skip reads {reads}!".format(reads=line))
                        else:
                            pass

            bam_in.close()

        return mrna_dict
    
    def read_bam(self):
        '''
        @Message  : import the mapping bam file
        @Input    : bam --> the mapping bam file
        @Return   : 
        @Flow     : step1 --> run
        '''
        
        """
        1. import the mapping bam file
        2. retrieve the reads and add them to gene message dictionary

        """
        if self.sample_format.lower() == ".bam":
            sys.stdout.writelines("import file: {bam}.\n".format(bam=self.bam))
            read_format = 'rb'
        elif self.sample_format.lower() == ".sam":
            sys.stdout.writelines("import file: {bam}.\n".format(bam=self.bam))
            read_format = 'r'
        else:
            sys.stdout.writelines("Unknown file format, please input the correct bam or sam file.")
            sys.exit()

        self.pysam_input = pysam.AlignmentFile(self.bam, read_format)

        # calculate the length distribution and eliminate the reads aligned to the negative strand.
        if self.profile == "monosome":
            self.monosome_p_site()
            # # run the flt_star_results function with multi-thread
            # gene_list = list(self.mrna_dict.keys())
            # gene_list_split = np.array_split(gene_list, self.thread)
            # gene_list_split = [list(split) for split in gene_list_split]

            # args = [(self.bam, self.mrna_dict, self.min_length, self.max_length, self.offset, 'rb') for i in range(self.thread)]

            # # run the format_rpf_density function with multi-thread
            # from multiprocessing import Pool
            # pool = Pool(processes=self.thread)
            # rpf_list = pool.map(self.flt_star_results, args)
            # pool.close()
            # pool.join()

            # # merge the long data from each sub-list
            # self.mrna_dict = pd.concat(rpf_list, ignore_index=True)

        elif self.profile == "disome":
            self.disome_p_site()
        elif self.profile == "trisome":
            self.trisome_p_site()

        self.pysam_input.close()
    
    @staticmethod
    def format_rpf_density(args):
        '''
        @Message  : format the rpf density to long data
        @Input    : gene_list --> the gene list
                    mrna_dict --> the gene message
                    out_prefix --> the output prefix
        @Return   : the long data in pandas dataframe format
        @Flow     : step1 --> retrieve the gene message from the rpf density
                    step2 --> check the 3nt periodicity and trim the utr length
                    step3 --> format the gene message to the long data
                    step4 --> format the rpf density to three columns
                    step5 --> return the long data
        '''

        gene_list, mrna_dict, out_prefix = args
        rpf_list = []

        for name in gene_list:
            # retrieve the gene message from the rpf density
            isoform = mrna_dict[name]
            rpf_dict = OrderedDict()

            # trim the cds frame to fit the 3nt periodicity
            if isoform.cds_length % 3 != 0:
                shift_cds = isoform.cds_length % 3
                isoform.cds_length -= shift_cds
                isoform.utr3_length += shift_cds

            shift_5 = isoform.utr5_length % 3
            shift_3 = isoform.utr3_length % 3

            utr5_length = isoform.utr5_length - shift_5
            utr3_length = isoform.utr3_length - shift_3
            trim_length = isoform.length - shift_5 - shift_3
            
            # format the gene message to the long data
            rpf_dict['name'] = [name] * int(trim_length / 3)
            rpf_dict['now_nt'] = list(range(1 + shift_5, isoform.length - shift_3, 3))
            rpf_dict['from_tis'] = list(range(-utr5_length // 3, (trim_length - utr5_length) // 3))
            rpf_dict['from_tts'] = list(range((utr3_length - trim_length) // 3 + 1, utr3_length // 3 + 1))
            rpf_dict['region'] = ['5utr'] * (utr5_length // 3) + ['cds'] * (isoform.cds_length // 3) + ['3utr'] * (utr3_length // 3)

            if shift_3 != 0:
                trim_seq = isoform.seq[shift_5: -shift_3]
                rpf_dict['codon'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

                trim_rpf = isoform.rpf[shift_5: -shift_3]
                rpf_dict[out_prefix + '_f0'] = trim_rpf[0::3]
                rpf_dict[out_prefix + '_f1'] = trim_rpf[1::3]
                rpf_dict[out_prefix + '_f2'] = trim_rpf[2::3]
            elif shift_3 == 0:
                trim_seq = isoform.seq[shift_5::]
                rpf_dict['codon'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

                trim_rpf = isoform.rpf[shift_5::]
                rpf_dict[out_prefix + '_f0'] = trim_rpf[0::3]
                rpf_dict[out_prefix + '_f1'] = trim_rpf[1::3]
                rpf_dict[out_prefix + '_f2'] = trim_rpf[2::3]
            try:
                rpf = pd.DataFrame(rpf_dict)
                rpf_list.append(rpf)
            except ValueError:
                print('Checking the input genes: ' + str(name), flush=True)

        total_rpf_df = pd.concat(rpf_list, ignore_index=True)
        return total_rpf_df

    def run_format_rpf_with_multi_thread(self):
        '''
        @Message  : run the format_rpf_density function with multi-thread
        @Input    : self.mrna_dict --> the gene message
                    self.thread --> the number of thread
        @Return   : self.total_rpf_df --> the long data in pandas dataframe format
        @Flow     : step1 --> split the gene list into several sub-list
                    step2 --> run the format_rpf_density function with multi-thread
                    step3 --> merge the long data from each sub-list
        '''

        gene_list = list(self.mrna_dict.keys())

        gene_list_split = np.array_split(gene_list, self.thread)
        gene_list_split = [list(split) for split in gene_list_split]

        args = [(gene_list_split[i], self.mrna_dict, self.output) for i in range(self.thread)]

        # run the format_rpf_density function with multi-thread
        from multiprocessing import Pool
        pool = Pool(processes=self.thread)
        rpf_list = pool.map(self.format_rpf_density, args)
        pool.close()
        pool.join()

        # merge the long data from each sub-list
        self.total_rpf_df = pd.concat(rpf_list, ignore_index=True)
    
        numeric_cols = self.total_rpf_df.select_dtypes(include='number').columns
        self.total_rpf_df[numeric_cols] = self.total_rpf_df[numeric_cols].astype(int)

    def output_density(self):
        
        total_reads_pl = pl.DataFrame(self.total_rpf_df)
        total_reads_pl.write_csv(self.output + "_rpf.txt", separator = '\t', include_header = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_CoV.py


from .ribo import ArgsParser
from .ribo import Coefficient_of_Variation


def main():
    ArgsParser.now_time()
    print('\nCalculate CoV in the CDS region.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_cov_args_parser()
    cov = Coefficient_of_Variation.CoV(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    cov.read_rpf()

    print('\nStep3: Calculate CoV.', flush=True)
    cov.cal_CoV()

    print('\nStep4: Output CoV.', flush=True)
    cov.output_CoV()

    if args.group:
        print('\nStep5: Compare CoV of each group.', flush=True)
        cov.read_group()
        cov.test_group_CoV()
    else:
        print('\nStep5: Skip CoV comparation.', flush=True)
        pass
    
    print('\nStep6: Draw CoV.', flush=True)
    cov.draw_fit_plot()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : find_peak.py


import time

from .ribo import ArgsParser
from .serp.SeRP import *


# main programme is here
def main():
    ArgsParser.now_time()
    
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.serp_peak_args_parser()
    serp = SeRP(args)
    serp.args_check()

    print('\nStep2: Import the RPF data.\n', flush=True)
    serp.rpf_txt_read()

    print('\nStep3: Import the gene annotation.\n', flush=True)
    serp.gene_anno()

    print('\nStep4: Scan peaks from the data.\n', flush=True)
    serp.detect_binding_peaks()

    print('\nStep5: Output peaks results.\n', flush=True)
    serp.output_peak()

    print("\nAll done!")
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Ensembl_Ref.py


from collections import OrderedDict


class Transcripts(object):

    def __init__(self, section):
        self.mess = [section["mess"]]

        self.chrom = section["chrom"]
        self.source = section["source"]
        self.feature = section["feature"]

        self.start = section["start"]
        self.end = section["end"]
        self.strand = section["strand"]
        self.attr = section["attr"]
        self.attr_dict = section["attr_dict"]

        self.gene_biotype = section["attr_dict"]["gene_biotype"]
        self.transcript_biotype = section["attr_dict"]["transcript_biotype"]
        self.gene_id = section["attr_dict"]["gene_id"]
        self.transcript_id = section["attr_dict"]["transcript_id"]
        self.transcript_name = None
        self.gene_name = None

        self.exon_feature = []
        self.cds_feature = []
        self.five_prime_utr = []
        self.three_prime_utr = []
        self.start_codon = []
        self.stop_codon = []
        self.exons = []
        self.cds = []
        self.utr5 = 0
        self.utr3 = 0
        self.cds_length = 0

        self.rep_transcript = False

        self.modified = False

        def get_ids():
            try:
                self.gene_name = section["attr_dict"]["gene_name"]
            except KeyError:
                self.gene_name = "-"
            try:
                self.transcript_name = section["attr_dict"]["transcript_name"]
            except KeyError:
                self.transcript_name = "-"

        get_ids()

    def add_feature(self, section, cds_shift):
        if section["feature"] == "exon":
            self.mess.append(section["mess"])
            self.exon_feature.append(section["mess"].split('\t'))
            self.exons.append([section["start"], section["end"]])

            if len(self.exons) >= 2 and section["start"] < self.exons[-2][1] and self.strand == "+":
                self.exon_feature.sort(key=lambda exon: exon[3], reverse=False)
                self.exons.sort(key=lambda exon: exon[0], reverse=False)

            elif len(self.exons) >= 2 and section["end"] > self.exons[-2][1] and self.strand == "-":
                self.exon_feature.sort(key=lambda exon: exon[3], reverse=True)
                self.exons.sort(key=lambda exon: exon[0], reverse=True)

        elif section["feature"] == "CDS":
            self.mess.append(section["mess"])
            self.cds_feature.append(section["mess"].split('\t'))
            self.cds.append([section["start"], section["end"]])
            self.cds_length += (section["end"] - section["start"] + cds_shift)

            if len(self.cds) >= 2 and section["start"] < self.cds[-2][1] and self.strand == "+":
                self.cds_feature.sort(key=lambda cds: cds[3], reverse=False)
                self.cds.sort(key=lambda cds: cds[0], reverse=False)

            elif len(self.cds) >= 2 and section["end"] > self.cds[-2][1] and self.strand == "-":
                self.cds_feature.sort(key=lambda cds: cds[3], reverse=True)
                self.cds.sort(key=lambda cds: cds[0], reverse=True)

        elif section["feature"] == "start_codon":
            self.start_codon = [section["start"], section["end"]]

        elif section["feature"] == "stop_codon":
            self.stop_codon = [section["start"], section["end"]]

    def check_codon(self):

        if self.strand == '+':
            if self.start_codon and self.start_codon[0] < self.cds[0][0]:
                if self.cds[0][0] - self.start_codon[0] > 3:
                    pass
                else:
                    self.cds_feature[0][3] = str(self.start_codon[0])
                self.cds[0][0] -= 3
                self.cds_length += 3
            if self.stop_codon and self.stop_codon[1] > self.cds[-1][1]:
                if self.stop_codon[1] - self.cds[-1][1] > 3:
                    pass
                else:
                    self.cds_feature[-1][4] = str(self.stop_codon[1])
                    self.cds[-1][1] += 3
                    self.cds_length += 3

        elif self.strand == '-':
            if self.start_codon and self.start_codon[1] > self.cds[0][1]:
                if self.start_codon[1] - self.cds[0][1] > 3:
                    pass
                else:
                    self.cds_feature[0][4] = str(self.start_codon[1])
                    self.cds[0][1] += 3
                    self.cds_length += 3
            if self.stop_codon and self.stop_codon[0] < self.cds[-1][0]:
                if self.cds[-1][0] - self.stop_codon[0] > 3:
                    pass
                else:
                    self.cds_feature[-1][3] = str(self.stop_codon[0])
                    self.cds[-1][0] -= 3
                    self.cds_length += 3

    def plus_utr5(self):
        self.utr5 = 0
        for exon in self.exons:
            if self.cds[0][0] > exon[1]:
                self.utr5 += exon[1] - exon[0] + 1
            elif exon[0] <= self.cds[0][0] <= exon[1]:
                self.utr5 += self.cds[0][0] - exon[0]

    def plus_utr3(self):
        self.utr3 = 0
        for exon in reversed(self.exons):
            if self.cds[-1][-1] < exon[0]:
                self.utr3 += exon[1] - exon[0] + 1
            elif exon[0] <= self.cds[-1][-1] <= exon[1]:
                self.utr3 += exon[1] - self.cds[-1][-1]

    def minus_utr5(self):
        self.utr5 = 0
        for exon in self.exons:
            if self.cds[0][-1] < exon[0]:
                self.utr5 += exon[1] - exon[0] + 1
            elif exon[0] <= self.cds[0][-1] <= exon[1]:
                self.utr5 += exon[1] - self.cds[0][-1]

    def minus_utr3(self):
        self.utr3 = 0
        for exon in reversed(self.exons):
            if self.cds[-1][0] > exon[1]:
                self.utr3 += exon[1] - exon[0] + 1
            elif exon[0] <= self.cds[-1][0] <= exon[1]:
                self.utr3 += self.cds[-1][0] - exon[0]

    def add_utr(self, utr_len, chroms_dict):
        self.check_codon()
        # check the plus strand
        if self.strand == '+':
            self.plus_utr5()
            self.plus_utr3()
            # weather the utr5 more than 30
            if self.utr5 < 90:
                if self.exons[0][0] - utr_len < 1:
                    self.exons[0][0] = 1
                    self.utr5 = self.cds[0][0] - 1
                    self.exon_feature[0][3] = 1
                    self.start = self.exons[0][0]
                else:
                    self.exons[0][0] = self.exons[0][0] - utr_len
                    self.utr5 = self.cds[0][0] - self.exons[0][0]
                    self.exon_feature[0][3] = self.exons[0][0]
                    self.start = self.exons[0][0]
                self.modified = utr_len
                self.plus_utr5()
            else:
                pass
            # weather the utr3 more than 30
            if self.utr3 < 90:
                if self.exons[-1][-1] + utr_len > chroms_dict[self.chrom].end:
                    self.exons[-1][-1] = chroms_dict[self.chrom].end
                    self.exon_feature[-1][4] = self.exons[-1][-1]
                    self.end = self.exons[-1][-1]
                else:
                    self.exons[-1][-1] = self.exons[-1][-1] + utr_len
                    self.exon_feature[-1][4] = self.exons[-1][-1]
                    self.end = self.exons[-1][-1]
                self.modified = utr_len
                self.plus_utr3()
            else:
                pass
        # 判断负链
        elif self.strand == '-':
            self.minus_utr5()
            self.minus_utr3()
            # weather the utr5 more than 30
            if self.utr5 < 90:
                if self.exons[0][-1] + utr_len > chroms_dict[self.chrom].end:
                    self.exons[0][-1] = chroms_dict[self.chrom].end
                    self.exon_feature[0][4] = self.exons[0][-1]
                    self.end = self.exons[0][-1]
                else:
                    self.exons[0][-1] = self.exons[0][-1] + utr_len
                    self.exon_feature[0][4] = self.exons[0][-1]
                    self.end = self.exons[0][-1]
                self.modified = utr_len
                self.minus_utr5()
            else:
                pass
            # weather the utr3 more than 30
            if self.utr3 < 90:
                if self.exons[-1][0] - utr_len < 1:
                    self.exons[-1][0] = 1
                    self.exon_feature[-1][3] = 1
                    self.start = 1
                else:
                    self.exons[-1][0] = self.exons[-1][0] - utr_len
                    self.exon_feature[-1][3] = self.exons[-1][0]
                    self.start = self.exons[-1][0]
                self.modified = utr_len
                self.minus_utr3()
            else:
                pass


class Gene(object):

    def __init__(self, section):
        self.section = section
        self.chrom = section["chrom"]
        self.source = section["source"]
        self.feature = section["feature"]
        self.gene_id = section["attr_dict"]["gene_id"]
        self.start = section["start"]
        self.end = section["end"]
        self.strand = section["strand"]
        self.attr = section["attr"]
        self.attr_dict = section["attr_dict"]

        self.gene_type = self.attr_dict["gene_biotype"]
        self.transcript = OrderedDict()
        self.rep_transcript = None
        self.rep_length = 0
        self.trans_num = 0
        self.modified = False

    def add_transcript(self, now_transcript):
        if now_transcript.cds_length > self.rep_length:
            self.rep_length = now_transcript.cds_length
            self.rep_transcript = now_transcript.transcript_id

        if now_transcript.modified:
            self.modified = True
            self.start = now_transcript.start
            self.end = now_transcript.end

        self.transcript[now_transcript.transcript_id] = now_transcript
        self.trans_num += 1


class Chrom(object):

    def __init__(self, line):
        self.attr = line.description
        self.chromosome = line.id
        self.start = 1
        self.end = len(line.seq)
        self.seq = line.seq



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_Offset_RSBM.py


from .ribo import ArgsParser
from .ribo.Offset_RSBM import *


def main():
    ArgsParser.now_time()
    sys.stdout.writelines('Detect the p-site offset.\n')
    sys.stdout.writelines('Step1: Checking the input Arguments.\n')
    args = ArgsParser.offset_rsbm_args_parser()
    offset_attr = Offset(args)

    sys.stdout.writelines('Step2: Import the transcripts annotation.\n')
    offset_attr.read_transcript()

    sys.stdout.writelines('Step3: Import the bam file.\n')
    offset_attr.get_mrna_reads()

    sys.stdout.writelines('Step4: Detect the RSBM offset of sequence profile.\n')
    offset_attr.get_frame_offset()
    offset_attr.format_frame_offset()
    offset_attr.adjust_frame_offset()

    sys.stdout.writelines('Step5: Output the frame offset.\n')
    offset_attr.write_frame_offset()

    sys.stdout.writelines('Step6: Draw the frame offset heatmap.\n')
    offset_attr.draw_frame_heatmap()

    sys.stdout.writelines('All done.\n')
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Project : riboParser
# @Script  : rand_seq.py


# import pandas as pd
# import polars as pl
import numpy as np
# from collections import OrderedDict
# from Bio import SeqIO
import argparse


def rand_seq_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to generate random sequence.')

    # needed arguments
    seq_class = parser.add_mutually_exclusive_group(required=True)
    seq_class.add_argument(
        '-d', dest='dna', required=False, action='store_true', default=False, 
        help='generate dna nucleotide sequence.'
    )
    seq_class.add_argument(
        '-r', dest='rna', required=False, action='store_true', default=False,
        help='generate rna nucleotide sequence.'
    )
    seq_class.add_argument(
        '-p', dest='protein', required=False, action='store_true', default=False,
        help='generate protein sequence.'
    )

    # optional arguments
    parser.add_argument(
        '-l', dest='length', required=False, type=int, default=10, 
        help='length of sequeuce. (default: %(default)s nt).'
    )
    parser.add_argument(
        '-n', dest='number', required=False, type=int, default=5, 
        help='number of sequence. (default: %(default)s).'
    )
    parser.add_argument(
        '-f', dest='format', required=False, choices=['fa', 'txt'], type=str, default='fa', 
        help='format of sequence. (default: %(default)s).'
    )

    args = parser.parse_args()
    # args_dict = vars(args)
    # for k, v in args_dict.items():
    #     print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def rand_seq(args):
    '''
    @Message  : function for generate the random sequence.
    @Input    : seq class --> dna/rna/protein
                seq length --> length of sequence
                seq number --> number of sequence
                seq format --> format of sequence

    @Return   : output --> random sequences
    @Flow     : step1 --> check the input arguments
                step2 --> generate the random sequence
                step3 --> output the random sequence
    '''
    
    # set the random seed
    # np.random.seed(1314)

    # check the sequence class
    if args.dna:
        seq_class = 'd'
        seq_list = ['A', 'T', 'C', 'G']

    elif args.rna:
        seq_class = 'r'
        seq_list = ['A', 'U', 'C', 'G']

    else:
        seq_class = 'p'
        seq_list = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 
                    'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']

    for i in range(args.number):
        seq = ''.join(np.random.choice(seq_list, args.length))
        if args.format == 'fa':
            print('>{}{}'.format(seq_class, i+1))
            print(seq)
        else:
            print(seq)


def main():
    args = rand_seq_args_parser()
    rand_seq(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_offset_end.py


import os
import pandas as pd
from collections import OrderedDict
import argparse


def merge_offset_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the details of tis offset files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for frame/tis offset files (e.g., '*5end.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='output file name.'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_offset_files(offset_list):
    '''
    @Message  : retrieve the offset table.
    @Input    : list --> pattern for reads offset files (e.g., '*end.txt')
    @Return   : 
                output --> output dataframe contain the reads offset
    @Flow     : step1 --> retrieve the reads offset
    '''

    offset_dict = OrderedDict()

    # for each gene file
    for offset_file in offset_list:
        file_name = os.path.basename(offset_file).split('.')[0]

        if file_name.endswith('tis_3end'):
            offset_model = 'tis_3end'
        elif file_name.endswith('tis_5end'):
            offset_model = 'tis_5end'
        elif file_name.endswith('tts_3end'):
            offset_model = 'tts_3end'
        elif file_name.endswith('tts_5end'):
            offset_model = 'tts_5end'
        else:
            print('Error: the file name is not correct, please check it.')
            exit()

        # import the offset file
        file_prefix = file_name.replace('_3end', '').replace('_5end', '').replace('_tis', '').replace('_tts', '')
        offset_df = pd.read_csv(offset_file, sep='\t', index_col=0)
        
        try:
            offset_dict[file_prefix][offset_model] = offset_df
        except KeyError:
            offset_dict[file_prefix] = {}
            offset_dict[file_prefix][offset_model] = offset_df

    return offset_dict


def process_offset_dict(offset_dict):
    '''
    @Message  : retrieve the offset table.
    @Input    : list --> pattern for reads offset files (e.g., '*end.txt')
    @Return   : 
                output --> output dataframe contain the reads offset
    @Flow     : step1 --> retrieve the reads offset
    '''

    offset_merge = pd.DataFrame()

    # for each gene file
    for file, offset in offset_dict.items():
        
        offset_df = pd.DataFrame()

        # for each offset model
        tis = pd.concat([offset['tis_5end'], offset['tis_3end']], axis=1)
        tts = pd.concat([offset['tts_5end'], offset['tts_3end']], axis=1)
        
        tis.index.name = 'Length'
        tts.index.name = 'Length'

        tis.reset_index(inplace=True)
        tts.reset_index(inplace=True)

        tis.insert(0, 'Site', 'TIS')
        tts.insert(0, 'Site', 'TTS')

        tts.insert(0, 'Sample', file)
        tis.insert(0, 'Sample', file)

        offset_df = pd.concat([tis, tts], axis=0)

        # merge the gene offset file
        if offset_merge.empty:
            offset_merge = offset_df
        else:
            offset_merge = pd.concat([offset_merge, offset_df], axis=0)

    return offset_merge


def output_table(offset_merge, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> dataframe contain the reads offset
    @Return   : output --> description
    @Flow     : step1 --> output the dataframe to txt file
    '''

    offset_merge.to_csv(output_file + '_offset_end.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_offset_args_parser()

    print('Step2: import the end of offset file.', flush=True)
    offset_dict = import_offset_files(args.list)
    offset_merge = process_offset_dict(offset_dict)

    print('Step3: output the end of offset table.', flush=True)
    output_table(offset_merge, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboparser
# @Script  : rpf_Bam_Filter.py


from .ribo import ArgsParser


def main():
    ArgsParser.now_time()
    print('Filter the specific length reads from bam file.\n', flush=True)
    print('Step1: Checking the input Arguments.', flush=True)
    args = ArgsParser.bam_filter_args_parser()

    from ribo import BamFilter
    bam_attr = BamFilter.BamFilter(args)

    print('\nStep2: Filter the alignment bam file.', flush=True)
    bam_attr.import_bam()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : ribotish_to_gtf.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/10/26 17:40:16
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import pandas as pd
# import polars as pl
# import numpy as np
from collections import OrderedDict
# from Bio import SeqIO
import argparse

def get_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to convert the ribotish result to genepred format.')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        '-i', dest='input', required=True, type=str, help='input file name.'
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output genepred file name.'
    )

    # optional arguments
    parser.add_argument(
        '-t', dest='type', required=False, type=str, choices=['all', 'smorf', 'annotated'], default='all', 
        help='select the ORF class. (default: %(default)s)'
    )
    parser.add_argument(
        '-p', dest='pvalue', required=False, type=float, default=0.05, help='threshold of the ribo pvalue. (default: %(default)s)'
    )
    parser.add_argument(
        '-l', dest='length', required=False, type=int, default=8, help='length of the amino acids. (default: %(default)s)'
    )
    parser.add_argument(
        '-c', dest='clean', required=False, action='store_true', default=False, help='drop all the fuzzy ORFs. (default: %(default)s)'
    )
    args = parser.parse_args()

    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def filter_the_pvalue(ribotish, args):
    '''
    @Message  : filter the pvalue and length of ORFs.
    @Input    : ribotish --> the ribotish result
    @Return   : ribotish --> filtered pvalue and length of ribotish results
    @Flow     : step1 --> filter the pvalue
                step2 --> filter the length of ORFs
    '''
    
    # filter the pvalue
    ribotish = ribotish.loc[ribotish['RiboPvalue'] <= args.pvalue, :]
    ribotish = ribotish.loc[ribotish['AALen'] >= args.length, :]

    return  ribotish


def filter_the_orf_class(ribotish, args):
    '''
    @Message  : filter the ORFs with different orf class.
    @Input    : ribotish --> filtered pvalue and length of ribotish results
    @Return   : ribotish --> ribotish results
    @Flow     : step1 --> filter the smORF and annotated genes
    '''
    
    # replace the ['5UTR', '3UTR', 'Novel'] string in the part of TisType to ['uORF', 'dORF', 'Novel']
    ribotish['TisType'] = ribotish['TisType'].str.replace("5'UTR", 'uORF')
    ribotish['TisType'] = ribotish['TisType'].str.replace("3'UTR", 'dORF')

    # clean the TisType
    if args.clean:
        # remove the TisType contain the string 'Known'
        ribotish = ribotish.loc[~ribotish['TisType'].str.contains('Known'), :]

        # remove the TisType contain the string 'CDSFrameOverlap'
        ribotish = ribotish.loc[~ribotish['TisType'].str.contains('CDSFrameOverlap'), :]
        
    else:
        pass

    # filter the smORF and annotated genes
    if args.type == 'smorf':
        # filter the TisType contain the string 5'UTR/3'UTR/Novel
        ribotish = ribotish.loc[ribotish['TisType'].str.contains("uORF|dORF|Novel"), :]

    elif args.type == 'annotated':
        # filter the TisType contain the string Annotated
        ribotish = ribotish.loc[ribotish['TisType'].str.contains("Annotated|Extended|Internal|Truncated"), :]

    else:
        pass

    return ribotish


def unique_the_orf(ribotish):
    '''
    @Message  : unique the ORFs with the same gene-blocks.
    @Input    : ribotish --> filtered ribotish results
    @Return   : ribotish --> unique ribotish results
    @Flow     : step1 --> sort the ORFs with genome position and length of amino acids
                step2 --> drop the ORFs with the same gene-blocks
    '''
    
    ribotish = ribotish.sort_values(by=['GenomePos', 'AALen'], ascending=[True, False])

    ribotish = ribotish.drop_duplicates(subset=['TisType', 'Blocks'], keep='first')

    return ribotish


def input_the_ribotish(args):
    '''
    @Message  : filter the pvalue and length of ORFs.
    @Input    : args --> input arguments
    @Return   : ribotish --> the ribotish result
    @Flow     : step1 --> input the ribotish result
                step2 --> drop the fuzzy ORFs
                step3 --> filter the pvalue and length of ORFs
                step4 --> filter the ORFs with different orf class
                step5 --> unique the ORFs with the same blocks
                step6 --> output the ribotish result
    '''

    ribotish = pd.read_csv(args.input, sep='\t', header=0, names=None)

    ribotish = filter_the_pvalue(ribotish, args)

    ribotish = filter_the_orf_class(ribotish, args)

    ribotish = unique_the_orf(ribotish)

    return ribotish


def retrieve_aa(ribotish, args):
    '''
    @Message  : retrieve the amino acid of the ORFs.
    @Input    : ribotish --> filtered ribotish results
    @Return   : aa_sequence --> the amino acid sequence of the ORFs
    @Flow     : step1 --> retrieve the amino acid sequence
                step2 --> format the amino acid sequence to fasta format
                step3 --> output the amino acid sequence
    '''

    with open(args.output + '.aa', 'w') as aa_output:
        for index, row in ribotish.iterrows():

            aa_name = row['TisType'] + '_' + row['Gid'] + '_' + row['Tid'] + '_' + row['GenomePos'] + ' ' + row['Blocks']
            aa_seq = row['AASeq'][:-1]

            aa_output.writelines('>' + aa_name + '\n')
            aa_output.writelines(aa_seq + '\n')
    

def retrieve_cds(ribotish, args):
    '''
    @Message  : retrieve the amino acid of the ORFs.
    @Input    : ribotish --> filtered ribotish results
    @Return   : aa_sequence --> the amino acid sequence of the ORFs
    @Flow     : step1 --> retrieve the cds sequence
                step2 --> format the cds sequence to fasta format
                step3 --> output the cds sequence
    '''

    with open(args.output + '.cds', 'w') as cds_output:
        for index, row in ribotish.iterrows():
            aa_name = row['Gid'] + '_' + row['Tid'] + ' ' + row['GenomePos'] + ' ' + row['TisType'] + ' ' + row['Blocks']
            aa_seq = row['Seq']

            cds_output.writelines('>' + aa_name + '\n')
            cds_output.writelines(aa_seq + '\n')


def get_cds_frame(strand, gene_blocks):
    '''
    @Message  : get the cds frame.
    @Input    : strand --> gene strand
                gene_blocks --> gene blocks
    @Return   : cds_frame --> frame of cds
    @Flow     : step1 --> check the different strand
                step2 --> get the list of gene blocks
                step3 --> calculate the length of cds
                step4 --> calculate the frame of cds
                
    @notes    : genepred --> cds length = txEnd - txStart
                gtf --> cds length = txEnd - txStart + 1
                bed --> cds length = txEnd - txStart

    '''

    if strand == '+':
        orf_length = 0

        for index, block in enumerate(gene_blocks):
            t_start, t_end = block.split('-')

            if index == 0:
                cds_frame = [0]
                exonStarts = [int(t_start)]
                exonEnds = [int(t_end)]

                continue

            exonStarts.append(int(t_start))
            exonEnds.append(int(t_end))
            cds_len = int(t_end) - int(t_start)
            
            orf_length += cds_len
            frame = orf_length % 3

            cds_frame = cds_frame + [frame]

    elif strand == '-':
        orf_length = 0
        
        exonEnds = []

        for index, block in enumerate(gene_blocks):
            t_start, t_end = block.split('-')

            if index == 0:
                cds_frame = [0]
                exonStarts = [int(t_start)]
                exonEnds = [int(t_end)]

                continue

            exonStarts = [int(t_start)] + exonStarts
            exonEnds = [int(t_end)] + exonEnds
            cds_len = int(t_end) - int(t_start)
            
            orf_length += cds_len
            frame = orf_length % 3

            cds_frame = [frame] + cds_frame

    else:
        pass
    
    cds_frame = ','.join([str(item) for item in cds_frame]) + ','
    exonStarts = ','.join([str(item) for item in exonStarts]) + ','
    exonEnds = ','.join([str(item) for item in exonEnds]) + ','

    return cds_frame, exonStarts, exonEnds


def retrieve_blocks(ribotish, args):
    '''
    @Message  : format the gene blocks to genepred.
    @Input    : ribotish --> filtered ribotish results
    @Return   : gene-blocks --> formatted gene-blocks
    @Flow     : step1 --> retrieve the gene-blocks
                step2 --> format the gene-blocks
                step3 --> output the gene-blocks file
    '''
    
    with open(args.output + '.genepred', 'w') as genepred_output:

        for index, row in ribotish.iterrows():
            # gene names
            name = row['TisType'] + '_' + row['Gid'] + '_' + row['Tid']
            chrom = row['GenomePos'].split(':')[0]
            # source = 'ribotish'
            # type = 'CDS'

            # gene position
            strand = row['GenomePos'].split(':')[-1]

            gene_blocks = row['Blocks'].split(',')
            exonCount = len(gene_blocks)

            exonStarts_list = [item.split('-')[0] for item in gene_blocks]
            exonEnds_list = [item.split('-')[-1] for item in gene_blocks]

            txStart = exonStarts_list[0]
            txEnd = exonEnds_list[-1]
            cdsStart = txStart
            cdsEnd = txEnd

            exonStarts = ','.join(exonStarts_list) + ','
            exonEnds = ','.join(exonEnds_list) + ','

            # output the gene-blocks
            # gene_blocks = [name, chrom, strand, txStart, txEnd, cdsStart, cdsEnd, exonCount, exonStarts, exonEnds, score, name2, cdsStartStat, cdsEndStat, IstringexonFrames]
            gene_blocks = [name, chrom, strand, txStart, txEnd, cdsStart, cdsEnd, exonCount, exonStarts, exonEnds]

            genepred_output.writelines('\t'.join([str(item) for item in gene_blocks]) + '\n')


def main():
    print('Convert ribotish results to genepred format.', flush=True)
    args = get_args_parser()

    print('Step1: input the ribotish.', flush=True)
    record = input_the_ribotish(args)

    print('Step2: output the cds file.', flush=True)
    retrieve_aa(record, args)

    print('Step3: output the amino acid file.', flush=True)
    retrieve_cds(record, args)

    print('Step4: output the genepred file.', flush=True)
    retrieve_blocks(record, args)

    print('Done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : test01.py
# @Script  : GenePred.py


import os
import stat
import subprocess
import sys
from bisect import bisect_left
from bisect import bisect_right
from collections import OrderedDict

import numpy as np
import pandas as pd
from Bio import SeqIO
from Bio import Seq


class Chrom(object):

    def __init__(self, line):
        self.attr = line.description
        self.chromosome = str(line.id)
        self.start = 1
        self.end = len(line.seq)
        self.seq = line.seq


class GenePred(object):
    def __init__(self, args):
        self.gene = OrderedDict()
        self.gtf = args.gtf
        self.gtf_format = os.path.splitext(self.gtf)[-1]
        self.gp_df = None
        self.columns = ['name', 'chrom', 'strand', 'txStart', 'txEnd', 'cdsStart', 'cdsEnd', 'exonCount',
                        'exonStarts', 'exonEnds', 'score', 'name2', 'cdsStartStat', 'cdsEndStat',
                        'IstringexonFrames']
        self.coding = args.coding
        self.longest = args.longest
        self.utr = args.utr

        # genome file
        self.fasta = args.genome
        self.chroms_dict = OrderedDict()
        self.mrna_seq = OrderedDict()
        self.cds_seq = OrderedDict()

        # output file
        self.out_prefix = args.output
        self.whole = args.whole
        self.gp_file = self.out_prefix + '.genepred'
        self.gtf_new = self.out_prefix + '.norm.gtf'
        self.whole_file = self.out_prefix + '.whole.txt'
        self.txt_file = self.out_prefix + '.norm.txt'
        self.rna_file = self.out_prefix + '.norm.rna.fa'
        self.cds_file = self.out_prefix + '.norm.cds.fa'

    def gtf2gp(self):
        '''
        @Message  : function to convert the gtf to genepred.
        @Input    : self.gtf_format --> format of the gtf file
                    self.gtf --> gtf file
                    self.gp_file --> genepred file

        @Return   : output --> description
        @Flow     : step1 --> check the gff3ToGenePred and gtfToGenePred in the current environment
                    step2 --> convert the gtf or gff to genepred
        '''
        
        script_path = os.path.split((os.path.abspath(sys.argv[0])))[0]
        nowpath = os.path.dirname(script_path)
        if self.gtf_format in ['.gff', '.gff3']:
            gff3ToGenePred = subprocess.run(['which', 'gff3ToGenePred'], capture_output=True, text=True)
            if gff3ToGenePred.returncode == 0:
                run_tmp = subprocess.run(['gff3ToGenePred', '-warnAndContinue', '-rnaNameAttr=attr',
                                        '-geneNameAttr=attr', self.gtf, self.gp_file])
            else:
                print('{0} not found in current environment.'.format('gff3ToGenePred'), flush=True)
                sys.exit(1)

        elif self.gtf_format == '.gtf':
            gtfToGenePred = subprocess.run(['which', 'gtfToGenePred'], capture_output=True, text=True)
            if gtfToGenePred.returncode == 0:
                run_tmp = subprocess.run(['gtfToGenePred', '-allErrors', '-genePredExt',
                                        '-ignoreGroupsWithoutExons', self.gtf, self.gp_file])
            else:
                print('{0} not found in current environment.'.format('gtfToGenePred'), flush=True)
                sys.exit(1)
        else:
            print('Unknown file format {0}.'.format(self.gtf_format),)
            print('Please input annotation file in GTF or GFF format.', flush=True)
            sys.exit(1)

    def gp2gtf(self):
        '''
        @Message  : function to convert the genepred to gtf file.
        @Input    : self.gp_file --> genepred file
                    self.gtf_new --> new gtf file
        @Return   : output --> genePred file
        @Flow     : step1 --> check the genePredToGtf in the current environment
                    step2 --> convert the genepred to gtf
        '''
        
        # script_path = os.path.split((os.path.abspath(sys.argv[0])))[0]
        # nowpath = os.path.dirname(script_path)

        genePredToGtf = subprocess.run(['which', 'genePredToGtf'], capture_output=True, text=True)

        if genePredToGtf.returncode == 0:
            run_tmp = subprocess.run(['genePredToGtf', 'file', '-utr', '-honorCdsStat',
                                      '-source=ribo', self.gp_file, self.gtf_new])
        else:
            print('{0} not found in current environment.'.format('genePredToGtf'), flush=True)
            sys.exit(1)
        # os.remove(self.gp_file)

    def read_genepred(self):
        '''
        @Message  : function import the parser the genepred file .
        @Input    : self.gp_file --> gene pred file derived from gtf file
        @Return   : self.gp_df --> dataframe of the genepred file
        @Flow     : step1 --> read the genepred file and rename the columns
                    step2 --> calculate the utr5, cds and utr3 length
        '''

        self.gp_df = pd.read_csv(self.gp_file, sep='\t', header=None)
        os.remove(self.gp_file)
        self.gp_df.columns = ['name', 'chrom', 'strand', 'txStart', 'txEnd', 'cdsStart', 'cdsEnd', 'exonCount',
                              'exonStarts', 'exonEnds', 'score', 'name2', 'cdsStartStat', 'cdsEndStat',
                              'IstringexonFrames']

        utr5_list = []
        cds_list = []
        utr3_list = []
        print('Import gtf annotation.', flush=True)

        for idx, rows in self.gp_df.iterrows():

            if idx % 1000 == 0:
                print('rows: ' + str(idx), flush=True)

            exonStarts = np.array(rows['exonStarts'].split(',')[:-1], dtype='int')
            exonEnds = np.array(rows['exonEnds'].split(',')[:-1], dtype='int')

            cdsStart = rows['cdsStart']
            cdsEnd = rows['cdsEnd']

            if rows['cdsStartStat'] == 'none' or rows['cdsEndStat'] == 'none':
                utr5_list.append(0)
                cds_list.append(0)
                utr3_list.append(0)

            elif rows['exonCount'] == 1:
                if rows.strand == '+':
                    utr5_list.append(int(abs(cdsStart - exonStarts)))
                    cds_list.append(int(abs(cdsEnd - cdsStart)))
                    utr3_list.append(int(abs(exonEnds - cdsEnd)))

                elif rows.strand == '-':
                    utr5_list.append(int(abs(exonEnds - cdsEnd)))
                    cds_list.append(int(abs(cdsEnd - cdsStart)))
                    utr3_list.append(int(abs(cdsStart - exonStarts)))

            elif rows['exonCount'] > 1:
                if rows.strand == '+':
                    intronStarts = exonStarts[1:]
                    intronEnds = exonEnds[:-1]
                    introns = intronStarts - intronEnds
                    intron_length = abs(sum(introns))

                    # get utr5 length
                    now_right = bisect_right(exonStarts, cdsStart) - 1
                    utr5_intron = sum(introns[0:now_right])
                    utr5_length = cdsStart - exonStarts[0] - utr5_intron

                    # get utr3 length
                    now_left = bisect_left(exonEnds, cdsEnd)
                    utr3_intron = sum(introns[now_left:])
                    utr3_length = exonEnds[-1] - cdsEnd - utr3_intron

                    # get cds length
                    cds_length = abs(cdsEnd - cdsStart) - intron_length + utr5_intron + utr3_intron

                    utr5_list.append(utr5_length)
                    cds_list.append(cds_length)
                    utr3_list.append(utr3_length)

                elif rows.strand == '-':
                    intronStarts = exonEnds[:-1]
                    intronEnds = exonStarts[1:]
                    introns = intronEnds - intronStarts
                    intron_length = abs(sum(introns))

                    # get utr5 length
                    now_right = bisect_right(exonStarts, cdsStart) - 1
                    utr3_intron = sum(introns[0:now_right])
                    utr3_length = cdsStart - exonStarts[0] - utr3_intron

                    # get utr3 length
                    now_left = bisect_left(exonEnds, cdsEnd)
                    utr5_intron = sum(introns[now_left:])
                    utr5_length = exonEnds[-1] - cdsEnd - utr5_intron
                    
                    # get cds length
                    cds_length = abs(cdsEnd - cdsStart) - intron_length + utr5_intron + utr3_intron

                    utr5_list.append(utr5_length)
                    cds_list.append(cds_length)
                    utr3_list.append(utr3_length)

            else:
                print('Error: gene range was wrong at :' + rows.name + '\n', flush=True)
                continue
        
        print('rows: ' + str(idx), flush=True)

        self.gp_df['utr5_length'] = utr5_list
        self.gp_df['cds_length'] = cds_list
        self.gp_df['utr3_length'] = utr3_list

    def get_rep_transcript_bak(self):
        '''
        @Message  : function to filter the representative transcripts.
        @Input    : self.gp_df --> dataframe of the genepred file
        @Return   : self.gp_df --> dataframe of the genepred file with annotated column [rep_transcript]
        @Flow     : step1 --> group the dataframe by gene name
                    step2 --> label the max cds length
                    step3 --> label the max utr5 + cds + utr3 length
                    step4 --> label the first transcript
                    step5 --> add the flag of representative transcript
        @Note     : low efficiency of this mothod to label the representative transcript.
        '''
        
        print('Filter representative transcripts.', flush=True)

        counter = 0
        def add_true_flag(group):

            nonlocal counter
            counter += 1
            if counter % 500 == 0:
                print('Representative gene: ' + str(counter), flush=True)

            if len(group) == 1:
                group['rep_transcript'] = True
                return group
            
            else:
                max_cds_length = group['cds_length'].max()
                max_utr5_cds_utr3 = group.loc[group['cds_length'] == max_cds_length, ['utr5_length', 'cds_length', 'utr3_length']].sum(axis=1).max()

                max_cds_index = group.loc[group['cds_length'] == max_cds_length].index
                max_utr5_cds_utr3_index = group.loc[group['utr5_length'] + group['cds_length'] + group['utr3_length'] == max_utr5_cds_utr3].index

                if len(max_cds_index) == 1:
                    group.loc[max_cds_index, 'rep_transcript'] = True
                elif len(max_utr5_cds_utr3_index) == 1:
                    group.loc[max_utr5_cds_utr3_index, 'rep_transcript'] = True
                else:
                    group.loc[max_cds_index[0], 'rep_transcript'] = True
                return group

        self.gp_df['rep_transcript'] = False
        self.gp_df = self.gp_df.groupby('name2').apply(add_true_flag)
        print('Representative gene: ' + str(counter), flush=True)

        # print(self.gp_df)

    def get_rep_transcript(self):
        '''
        @Message  : function to filter the representative transcripts.
        @Input    : self.gp_df --> dataframe of the genepred file
        @Return   : self.gp_df --> dataframe of the genepred file with annotated column [rep_transcript]
        @Flow     : step1 --> calculate the gene length, length = cds_length + utr5_length + utr3_length
                    step2 --> sort the dataframe by gene name, cds length, gene length, utr5 length and utr3 length
                    step3 --> reset the index of the dataframe
                    step4 --> group the dataframe by gene name and label the max cds length and max gene length
                    step5 --> label the first transcript
                    step6 --> reset the index of the dataframe
        '''

        print('Filter representative transcripts.', flush=True)

        self.gp_df['gene_length'] = self.gp_df['cds_length'] + self.gp_df['utr5_length'] + self.gp_df['utr3_length']
        self.gp_df = self.gp_df.sort_values(by=['name2', 'cds_length', 'gene_length', 'utr5_length', 'utr3_length'], 
                                            ascending=[False, False, False, False, False])

        self.gp_df.reset_index(drop=True, inplace=True)
        self.gp_df['rep_transcript'] = False
        self.gp_df.loc[self.gp_df.groupby('name2').head(1).index, 'rep_transcript'] = True
        self.gp_df.sort_values(by=['chrom', 'txStart', 'name2', 'rep_transcript'], 
                               ascending=[True, True, True, False],
                               inplace=True)
        self.gp_df.reset_index(drop=True, inplace=True)

    def add_utr(self):
        '''
        @Message  : function to add the pseudo utr.
        @Input    : self.gp_df --> dataframe of the genepred file
        @Return   : self.gp_df --> dataframe of the genepred file with annotated column [modified]
        @Flow     : step1 --> create the list to store the new exon starts and ends
                    step2 --> skip the transcript without cds
                    step3 --> calculate the new exon start site to extend the utr5
                    step4 --> calculate the new exon end site to extend the utr3
                    step5 --> add the pseudo utr to the transcript
        '''
        
        exonStarts_list, exonEnds_list = [], []
        txStart_list, txEnd_list = [], []
        utr5_list, utr3_list = [], []

        self.gp_df.loc[:, 'modified'] = 'False'

        # add pseudo utr
        if self.utr == 0:
            pass

        else:
            print('Add pseudo UTR.', flush=True)

            for idx, rows in self.gp_df.iterrows():
                if idx % 1000 == 0:
                    print('rows: ' + str(idx), flush=True)

                # skip the transcript without cds
                if rows['cdsStartStat'] == 'none' or rows['cdsEndStat'] == 'none':
                    exonStarts_list.append(rows.exonStarts)
                    exonEnds_list.append(rows.exonEnds)
                    txStart_list.append(rows.txStart)
                    txEnd_list.append(rows.txEnd)
                    utr5_list.append(rows.utr5_length)
                    utr3_list.append(rows.utr3_length)
                    continue

                # check exon starts for each genes
                exonStarts = rows['exonStarts'].split(',')[:-1]
                
                if int(rows['utr5_length']) >= self.utr:
                    # skip the transcript with utr5 length >= self.utr
                    exonStarts_list.append(rows.exonStarts)
                    txStart_list.append(rows.txStart)
                    utr5_list.append(rows.utr5_length)

                else:
                    # calculate the new exon start site to extend the utr5
                    new_start = int(exonStarts[0]) - self.utr

                    if new_start < 0:
                        # set the negative new start site to 0, since the chromosome start site can not be negative
                        exonStarts[0] = '0'
                        exonStarts_list.append(','.join(exonStarts) + ',')
                        txStart_list.append(0)
                        utr5_list.append(rows['cdsStart'])
                        self.gp_df.loc[idx, 'modified'] = 'True'
                    else:
                        # set the new start site to new_start
                        exonStarts[0] = str(new_start)
                        exonStarts_list.append(','.join(exonStarts) + ',')
                        txStart_list.append(rows.txStart - self.utr)
                        utr5_list.append(rows.utr5_length + self.utr)
                        self.gp_df.loc[idx, 'modified'] = 'True'

                # check exon ends for each genes
                exonEnds = rows['exonEnds'].split(',')[:-1]

                if int(rows['utr3_length']) >= self.utr:
                    # skip the transcript with utr3 length >= self.utr
                    exonEnds_list.append(rows.exonEnds)
                    txEnd_list.append(rows.txEnd)
                    utr3_list.append(rows.utr3_length)
                else:
                    # calculate the new exon end site to extend the utr3
                    new_end = int(exonEnds[-1]) + self.utr

                    if new_end > self.chroms_dict[str(rows['chrom'])].end:
                        # set the new end site to chromosome end site, since the chromosome end site can not be larger than the chromosome end site
                        exonEnds[-1] = str(self.chroms_dict[str(rows['chrom'])].end)
                        exonEnds_list.append(','.join(exonEnds) + ',')
                        txEnd_list.append(self.chroms_dict[str(rows['chrom'])].end)
                        utr3_list.append(rows.utr3_length + self.chroms_dict[str(rows['chrom'])].end - rows['cdsEnd'])
                        self.gp_df.loc[idx, 'modified'] = 'True'
                    else:
                        # set the new end site to new_end
                        exonEnds[-1] = str(new_end)
                        exonEnds_list.append(','.join(exonEnds) + ',')
                        txEnd_list.append(new_end)
                        utr3_list.append(rows.utr3_length + self.utr)
                        self.gp_df.loc[idx, 'modified'] = 'True'

            self.gp_df['exonStarts'] = exonStarts_list
            self.gp_df['exonEnds'] = exonEnds_list
            self.gp_df['txStart'] = txStart_list
            self.gp_df['txEnd'] = txEnd_list
            self.gp_df['utr5_length'] = utr5_list
            self.gp_df['utr3_length'] = utr3_list

    def read_genome(self):
        '''
        @Message  : function import the genome sequences.
        @Input    : self.fasta --> genome file
        @Return   : self.chroms_dict --> dictionary of the chromosome sequences
        @Flow     : step1 --> read the genome file with SeqIO
                    step2 --> store the chromosome sequences in the dictionary
        '''
        
        record = SeqIO.parse(self.fasta, "fasta")

        for line in record:
            print("import chromosome: {chrom}".format(chrom=line.id), flush=True)
            self.chroms_dict[str(line.id)] = Chrom(line)


    def get_seq(self):
        '''
        @Message  : function to retrieve the mRNA sequences.
        @Input    : self.gp_df --> dataframe of the genepred file
        @Return   : self.mrna_seq --> dictionary of the mRNA sequences
        @Flow     : step1 --> iterate the dataframe to get the mRNA sequences
                    step2 --> store the mRNA sequences in the dictionary
                    step3 --> store the cds sequences in the dictionary
        '''

        for idx, rows in self.gp_df.iterrows():

            # skip the transcript without cds
            if rows['cds_length'] == 0:
                continue

            chromesome = str(rows['chrom'])
            rna = list(zip(rows.exonStarts.split(',')[:-1], rows.exonEnds.split(',')[:-1]))
            tmp_rna_seq = Seq.Seq('')

            if rows['strand'] == "-":
                for exon in rna:
                    exon_start, exon_end = int(exon[0]), int(exon[1])
                    tmp_rna_seq += self.chroms_dict[chromesome].seq[exon_start: exon_end]
                tmp_rna_seq = tmp_rna_seq.reverse_complement()

            else:
                for exon in rna:
                    exon_start, exon_end = int(exon[0]), int(exon[1])
                    tmp_rna_seq += self.chroms_dict[chromesome].seq[exon_start: exon_end]

            self.mrna_seq[rows['name']] = tmp_rna_seq
        
            self.cds_seq[rows['name']] = tmp_rna_seq[rows['utr5_length']: rows['utr5_length'] + rows['cds_length']]

    def write_txt(self):
        '''
        @Message  : function to output the txt file.
        @Input    : self.gp_df --> dataframe of the genepred file
        @Return   : gp_df_new --> dataframe of the genepred file
                    txt_df --> dataframe of the txt file
        @Flow     : step1 --> output the whole genepred file
                    step2 --> filter the coding transcript
                    step3 --> filter the longest transcript
                    step4 --> output the genePred file
                    step5 --> format the genePred file to txt file
        '''
        
        if self.whole:
            self.gp_df.to_csv(self.whole_file, sep='\t', header=True, index=None)

        if self.coding:
            self.gp_df = self.gp_df.loc[self.gp_df['cds_length'] != 0, ]

        if self.longest:
            self.gp_df = self.gp_df.loc[self.gp_df['rep_transcript'] == 'True', ]

        self.gp_df.loc[:, 'name'] = self.gp_df.loc[:, 'name'].str.replace('cds-', '').str.replace('rna-', '')
        self.gp_df.loc[:, 'name2'] = self.gp_df.loc[:, 'name2'].str.replace('gene-', '')

        # output the genePred file
        gp_df_new = self.gp_df.loc[:, self.columns].copy()
        gp_df_new.to_csv(self.gp_file, sep='\t', header=None, index=None)

        # output the txt file
        txt_columns = ['chrom', 'name2', 'name', 'txStart', 'txEnd', 'utr5_length', 'cds_length', 'utr3_length',
                       'strand', 'rep_transcript', 'modified']
        txt_df = self.gp_df.loc[:, txt_columns].copy()

        txt_df.columns = ['chromosome', 'gene_id', 'transcript_id', 'start', 'end', 'utr5_length',
                          'cds_length', 'utr3_length', 'strand', 'rep_transcript', 'modified']
        
        txt_df.loc[:, ['utr5_length', 'cds_length', 'utr3_length']] = txt_df.loc[:, ['utr5_length', 'cds_length', 'utr3_length']].astype(int)
        txt_df = txt_df.loc[txt_df['cds_length'] != 0, ]
        txt_df.to_csv(self.txt_file, sep='\t', header=True, index=None)

    def write_seq(self):
        '''
        @Message  : function to output the mRNA and cds sequences.
        @Input    : self.rna_file --> mrna sequences file
                    self.cds_file --> cds sequences file
        '''
        
        with open(self.rna_file, 'w') as seq_out:
            for rna, mess in self.mrna_seq.items():
                seq_out.writelines('\n'.join(['>' + rna, str(mess)]) + '\n')

        with open(self.cds_file, 'w') as cds_out:
            for cds, mess in self.cds_seq.items():
                cds_out.writelines('\n'.join(['>' + cds, str(mess)]) + '\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_density.py


from .ribo import ArgsParser
from .ribo import Ribo


def main():
    ArgsParser.now_time()
    print('\nConvert reads to RPFs density.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.ribo_args_parser()

    ribo_attr = Ribo.Ribo(args)

    print('\nStep2: Import the P-site offset.\n', flush=True)
    ribo_attr.read_offset()

    print('\nStep3: Import the transcripts annotation.', flush=True)
    ribo_attr.read_transcript()

    print('\nStep4: Import the BAM file.\n', flush=True)
    ribo_attr.read_bam()

    print('\nStep5: Format the in-frame RPFs density.', flush=True)
    ribo_attr.run_format_rpf_with_multi_thread()

    print('\nStep5: Output the RPFs density.', flush=True)
    ribo_attr.output_density()

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_cst.py


import pandas as pd
import argparse


def merge_cst_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the codon decoding time files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for cst files (e.g., '*_codon_selection_time.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name (default: prefix + _codon_selection_time.txt).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_cst_files(cst_list):
    '''
    @Message  : retrieve the cst table.
    @Input    : list --> pattern for cst files (e.g., '*_codon_selection_time.txt')
    @Return   : 
                output --> output dataframe contain the cst
    @Flow     : step1 --> retrieve the cst
    '''

    # make the dataframe
    cst_df = pd.DataFrame()

    # for each cst file
    for cst_file in cst_list:
        # import the cst file
        cst = pd.read_csv(cst_file, sep='\t', index_col=False)
        cst.sort_values(by=['Abbr', 'Codon'], ascending=True, inplace=True, ignore_index=True)

        # merge the gene cst file
        if cst_df.empty:
            cst_df = cst
        else:
            # delete the columns ['AA', 'Abbr.']
            cst = cst.drop(['AA', 'Abbr'], axis=1)
            cst_df = pd.merge(left=cst_df, right=cst, how='outer', on='Codon')

    # reorder the cst columns with 
    # ['codon', 'AA', 'Abbr.', *_density, *_absolute_cst, *_relative_cst]
    cst_codon = cst_df[['Codon', 'AA', 'Abbr']]
    absolute_cst = cst_df.filter(regex='_absolute_cst')
    relative_cst = cst_df.filter(regex='_relative_cst')

    cst_merge = pd.concat([cst_codon, absolute_cst, relative_cst], axis=1)

    return cst_merge


def output_table(cst_merge, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the cst
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    cst_merge.to_csv(output_file + '_codon_selection_time.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_cst_args_parser()

    print('Step2: import the cst file.', flush=True)
    cst_df = import_cst_files(args.list)

    print('Step3: output the cst table.', flush=True)
    output_table(cst_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_metagene.py


import os
import pandas as pd
from collections import OrderedDict
import argparse


def merge_metagene_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the metagene files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for metagene files (e.g., '*_metaplot.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name (default: prefix + _metaplot.txt).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_metagene_files(metagene_list):
    '''
    @Message  : retrieve the metagene table.
    @Input    : list --> pattern for metagene files (e.g., '*tis_tts_metaplot.txt')
    @Return   : 
                output --> output dataframe contain the metagene
    @Flow     : step1 --> retrieve the metagene
    '''

    # make the dataframe
    metagene_colunms = ['Sample', 'Meta', 'Nucleotide', 'Codon', 'Frame', 'Density']
    metagene_df = pd.DataFrame(columns=metagene_colunms)

    # for each gne file
    for metagene_file in metagene_list:

        # import the metagene file
        metagene = pd.read_csv(metagene_file, sep='\t', index_col=False)

        # merge the gene metagene file
        metagene_df = pd.concat([metagene_df, metagene], axis=0)

    return metagene_df


def output_table(metagene_df, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the metagene
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    metagene_df.to_csv(output_file + '_metagene.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_metagene_args_parser()

    print('Step2: import the metagene file.', flush=True)
    metagene_df = import_metagene_files(args.list)

    print('Step3: output the metagene table.', flush=True)
    output_table(metagene_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Project : riboParser
# @Script  : Cumulative_CoV.py


# import pandas as pd
# import polars as pl
# import numpy as np
# from collections import OrderedDict
# from Bio import SeqIO
# import argparse

import pandas as pd
from . import RPFs


class CumulativeCoV(object):

    def __init__(self, args):
        # input and output
        self.rpf = args.rpf
        self.out = args.output

        self.list = args.list
        self.gene_rpf = None

        # parameters for rpfs filter
        self.rpf_num = args.min
        self.trim = args.trim

        self.norm = args.normal
        self.zero = args.zero

        # format rpf table
        self.split = args.split
        self.format = True
        
        # rpf table
        self.sample_name = None
        self.sample_num = None
        self.total_rpf_num = None
        self.raw_rpf = None


    def retrieve_rpf(self):
        '''
        @Message  : import the rpf table.
        @Input    : self.rpf --> rpf table
                    self.site --> [E, P, A] site in ribosome
                    self.frame --> [0, 1, 2] frame in [E, P, A] site
                    self.gene --> gene list
                    self.rpf_num --> minimum rpf value
                    norm --> normalize the rpf value to RPM
        @Return   : self.raw_rpf --> rpf table contain all rpf value
                    self.sample_name --> samples name
                    self.sample_num --> groups sample number
                    self.merged_rpf --> merged rpf table
                    self.total_rpf_num --> total rpf number of each sample
        '''
        
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sample_num=None,
                                      sample_name=None,
                                      sites='P',
                                      frame='all',
                                      gene=self.list,
                                      rpf_num=self.rpf_num)

        self.raw_rpf = rpf_results[0].to_pandas()
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]

        self.high_rpf = self.raw_rpf.loc[self.raw_rpf['name'].isin(self.high_gene), ]

        del rpf_results


    def rpf_to_rpm(self):
        '''
        @Message  : convert rpf to rpm value.
        @Input    : high_rpf --> rpf table in pandas dataframe
                    norm --> default [False]
        @Return   : high_rpf --> rpm table
        '''
        
        if self.norm:
            total_rpf_num = [x for x in self.total_rpf_num for _ in range(3)]
            self.high_rpf.iloc[:, 6::] = self.high_rpf.iloc[:, 6:] * 1000000 / total_rpf_num
        else:
            pass

        self.high_rpf = self.high_rpf[self.high_rpf['region'] == 'cds']

        # set the cds start position to 1
        if self.zero:
            self.high_rpf['now_nt'] = self.high_rpf.groupby('name')['now_nt'].transform(lambda x: x - x.min() + 1)


    def melt_rpf_table(self):
        '''
        @Message  : melt the data into a format that is convenient for drawing.
        @Input    : self.high_rpf --> specific genes rpf table 
        @Return   : self.melt_rpf --> melt the three frame data to long data format

        from
            name ... wt_f0 wt_f1 wt_f2
            ORF1 ... 10 3   2
            ORF1 ... 7 2   2
        to
            name ... wt
            ORF1 ... 10
            ORF1 ... 3
            ORF1 ... 2
            ...
        '''

        ids_vars = self.high_rpf.columns[0:6].to_list()
        sample_title = [x + y for x in self.sample_name for y in ['_f0', '_f1', '_f2']]
        self.gene_rpf = pd.melt(self.high_rpf, id_vars=ids_vars, value_vars=sample_title, var_name='sample', value_name='rpf')
        
        self.gene_rpf['frame'] = self.gene_rpf['sample'].str[-1]
        self.gene_rpf['sample'] = self.gene_rpf['sample'].str[:-3]

        self.gene_rpf = self.gene_rpf.sort_values(by=['sample', 'name', 'now_nt', 'frame']).reset_index(drop=True)
        self.gene_rpf = self.gene_rpf.pivot(index=ids_vars + ['frame'], columns='sample', values='rpf').reset_index()
        self.gene_rpf['now_nt'] = self.gene_rpf['now_nt'].astype(int) + self.gene_rpf['frame'].astype(int)
        self.gene_rpf['codon'] = self.gene_rpf.apply(lambda row: row['codon'][int(row['frame'])], axis=1)


    def calc_cov(self):
        '''
        @Message  : calculate the coefficient of variation.
        @Input    : self.gene_rpf --> rpf table
        @Return   : self.gene_cov --> coefficient of variation table

        from

        gene	site	a	b
        x	1	4	5
        x	2	5	5
        x	3	10	1
        x	4	3	0
        x	5	10	3
        x	6	1	4
        x	7	9	1
        x	8	4	2
        x	9	7	3
        x	10	2	2
        y	1	5	3
        y	2	2	1
        y	3	3	4
        y	4	4	4
        y	5	5	1
        y	6	6	2
        y	7	4	2
        y	8	6	4
        y	9	7	1
        y	10	5	3
        y	11	3	0
        y	12	5	1

        to

        gene	site	a	b
        x	1	nan	0
        x	2	0.15713484	0
        x	3	0.507560566	0.514259477
        x	4	0.5652957	0.828221234
        x	5	0.52524176	0.728431359
        x	6	0.677867341	0.638284739
        x	7	0.608580619	0.702192845
        x	8	0.600656765	0.685118789
        x	9	0.553155291	0.637377439
        x	10	0.601497977	0.624926031
        y	1	0	0
        y	2	0.333333333	0.428571429
        y	3	0.40824829	0.374165739
        y	4	0.447213595	0.319438282
        y	5	0.471404521	0.306892205
        y	6	0.487950036	0.32249031
        y	7	0.5	0.30061372
        y	8	0.509175077	0.301018679
        y	9	0.516397779	0.319438282
        y	10	0.522232968	0.301647806
        y	11	0.527046277	0.316227766
        y	12	0.531085005	0.301511345

        '''
        
        # group the data by gene name
        gene_group = self.gene_rpf.groupby('name')

        # calculate the coefficient of variation
        def calc_cov(df, column):
            return df[column].expanding().std() / df[column].expanding().mean()

        # calculate the coefficient of variation for each gene
        gene_cov = gene_group.apply(calc_cov, self.sample_name)

        gene_cov = gene_cov.reset_index()

        self.gene_rpf[self.sample_name] = gene_cov[self.sample_name]


    def merge_cov_table(self):
        '''
        @Message  : merge the coefficient of variation table
        @Input    : self.gene_rpf --> rpf table
                    self.gene_cov --> coefficient of variation table
        '''
        # delete the gene with max(now_nt) < self.trim
        gene_rpf_list = self.gene_rpf[self.gene_rpf['now_nt'] > self.trim]['name'].unique().tolist()

        # trim the now_nt < self.trim
        gene_rpf_trim = self.gene_rpf[self.gene_rpf['name'].isin(gene_rpf_list)]
        gene_rpf_trim = gene_rpf_trim[self.gene_rpf['now_nt'] <= self.trim]
        gene_rpf_trim = gene_rpf_trim[['now_nt'] + self.sample_name]

        gene_rpf_meta = gene_rpf_trim.groupby('now_nt').mean().reset_index()
        gene_rpf_meta = pd.melt(gene_rpf_meta, id_vars='now_nt', value_vars=self.sample_name, 
                                var_name='Sample', value_name='Density')

        gene_rpf_meta['Frame'] = gene_rpf_meta['now_nt'].apply(lambda x: (x - 1) % 3 )
        gene_rpf_meta['Codon'] = gene_rpf_meta['now_nt'].apply(lambda x: (x - 1) // 3 + 1)
        gene_rpf_meta['Nucleotide'] = gene_rpf_meta['now_nt']
        gene_rpf_meta['Meta'] = "TIS"
        
        gene_rpf_meta = gene_rpf_meta[['Sample', 'Meta', 'Nucleotide', 'Codon', 'Frame', 'Density']]

        gene_rpf_meta.to_csv(self.out + "_meta" + str(self.trim) + "_cumulative_CoV.txt", sep="\t", index=False)


    def output_rpf_table(self):
        '''
        @Message  : split merged rpf table to each file.
        @Input    : self.melt_rpf --> description
        @Return   : output --> output rpf table in txt format
        '''
        
        if self.split:
            gene_group = self.gene_rpf.groupby('name')
            for gene, expr in gene_group:
                filename = gene
                expr.to_csv(filename + "_cumulative_CoV.txt", sep="\t", index=False)
        else:
            self.gene_rpf.to_csv(self.out + "_cumulative_CoV.txt", sep="\t", index=False)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : ribocode_bed_format.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/10/27 15:02:11
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import pandas as pd
# import polars as pl
# import numpy as np
from collections import OrderedDict
# from Bio import SeqIO
import argparse
import copy


def get_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to convert the ribocode bed file to genepred format.')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        '-i', dest='input', required=True, type=str, help='input file name.'
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output genepred file name.'
    )

    # optional arguments
    parser.add_argument(
        '-t', dest='type', required=False, type=str, choices=['all', 'smorf', 'annotated'], default='all', 
        help='select the ORF class. (default: %(default)s)'
    )
    parser.add_argument(
        '-l', dest='length', required=False, type=int, default=8, help='length of the amino acids. (default: %(default)s)'
    )
    parser.add_argument(
        '-c', dest='clean', required=False, action='store_true', default=False, help='drop all the fuzzy ORFs. (default: %(default)s)'
    )
    args = parser.parse_args()

    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def filter_the_orf_class(ribocode, args):
    '''
    @Message  : filter the ORFs with different orf class.
    @Input    : ribocode --> filtered pvalue and length of ribocode results
    @Return   : ribocode --> ribocode results
    @Flow     : step1 --> filter the smORF and annotated genes

    '''
    
    tmp_ribocode = copy.deepcopy(ribocode)

    # filter the ORFType
    if args.type == 'smorf':
        # filter the smORF genes
        for index, orf in tmp_ribocode.items():
            if orf[-1] not in ['uORF', 'dORF', 'novel', 'dORF,novel', 'uORF,novel']:
                del ribocode[index]
        
    elif args.type == 'annotated':
        # filter the ORFType contain the string Annotated
        for index, orf in tmp_ribocode.items():
            if orf[-1] not in ['internal', 'annotated']:
                del ribocode[index]

    else:
        pass
    
    return ribocode


def flt_the_orf_length(ribocode, args):
    '''
    @Message  : unique the ORFs with the same gene-blocks.
    @Input    : ribocode --> filtered ribocode results
    @Return   : ribocode --> unique ribocode results
    @Flow     : step1 --> sort the ORFs with genome position and length of amino acids
                step2 --> drop the ORFs with the same gene-blocks
    '''
    
    tmp_ribocode = copy.deepcopy(ribocode)

    for index, orf in tmp_ribocode.items():
        orf_length = int(index.split('_')[-1])

        if orf_length < args.length:
            del ribocode[index]

    return ribocode


def fix_stop_codon(ribocode):
    '''
    @Message  : fix the stop codon of the ORFs.
    @Input    : ribocode --> filtered ribocode results
    @Return   : ribocode --> fixed stop codon of ORFs
    @Flow     : step1 --> fix the stop codon of the ORFs
    '''

    for index, orf in ribocode.items():
        if orf[2] == '+':
            orf[9][-1] = str(int(orf[9][-1]) + 3)

        else:
            orf[8][0] = str(int(orf[8][0]) - 3)

        orf[3] = orf[8][0]
        orf[4] = orf[9][-1]
        orf[5] = orf[8][0]
        orf[6] = orf[9][-1]

    return ribocode


def input_the_ribocode(args):
    '''
    @Message  : filter the pvalue and length of ORFs.
    @Input    : args --> input arguments
    @Return   : ribocode --> the ribocode result
    @Flow     : step1 --> input the ribocode result
                step2 --> drop the fuzzy ORFs
                step3 --> filter the length of ORFs
                step4 --> filter the ORFs with different orf class
                step5 --> output the ribocode result
    '''

    # input the ribocode result
    ribocode = OrderedDict()

    with open(args.input, 'r') as ribocode_input:
        for line in ribocode_input:
            if line.startswith('#'):
                continue
            
            record = line.strip().split('\t')
            
            # orf position
            chrom = record[0]

            # orf name
            orf_mess = record[3].split(';')

            orf_name = orf_mess[0]
            gene = orf_mess[1]
            rna = orf_mess[2]
            orf_type = orf_mess[3]

            # orf position
            txStart = int(orf_name.split('_')[-3])
            txEnd = int(orf_name.split('_')[-2])

            # orf strand
            score = record[4]
            strand = record[5]

            # fix the stop codon to the CDS
            if strand == '-':
                txStart, txEnd = txEnd - 1, txStart
            else:
                txStart, txEnd = txStart - 1, txEnd

            cdsStart = txStart
            cdsEnd = txEnd

            # fix the edge of the orf
            exonStarts = [record[1]]
            exonEnds = [record[2]]

            # store the orf information
            if orf_name not in ribocode.keys():
                exonCount = 1
                name = orf_type + '_' + orf_name

                ribocode[orf_name] = [name, chrom, strand, txStart, txEnd, cdsStart, cdsEnd, exonCount, exonStarts, exonEnds, orf_type]
                
            else:
                exonCount = ribocode[orf_name][7] + 1
                if strand == '-':
                    exonStarts = exonStarts + ribocode[orf_name][8]
                    exonEnds = exonEnds + ribocode[orf_name][9]
                else:
                    exonStarts = ribocode[orf_name][8] + exonStarts
                    exonEnds = ribocode[orf_name][9] + exonEnds

                ribocode[orf_name] = [name, chrom, strand, txStart, txEnd, cdsStart, cdsEnd, exonCount, exonStarts, exonEnds, orf_type]

    ribocode = filter_the_orf_class(ribocode, args)

    ribocode = flt_the_orf_length(ribocode, args)

    ribocode = fix_stop_codon(ribocode)

    return ribocode


def output_genepred(ribocode, args):
    '''
    @Message  : format the gene blocks to genepred.
    @Input    : ribocode --> filtered ribocode results
    @Return   : gene-blocks --> formatted gene-blocks
    @Flow     : step1 --> convert the gene block list to string
                step2 --> convert dict to dataframe, and sort the chrom and start position
                step3 --> output the gene-pred file
    '''
    
    # convet list to string
    for index, orf in ribocode.items():
        orf[8] = ','.join(orf[8]) + ','
        orf[9] = ','.join(orf[9]) + ','

    # convert dict to dataframe
    ribocode = pd.DataFrame.from_dict(ribocode, orient='index')
    ribocode[3] = ribocode[3].astype(int)

    # sort the chrom and start position
    ribocode = ribocode.sort_values(by=[1, 3], ascending=[True, True])

    # delete the last column of the ribocode 
    ribocode = ribocode.iloc[:, :-1]

    # output the genepred file
    ribocode.to_csv(args.output + '.genepred', sep='\t', header=False, index=False)


def main():
    print('Convert ribocode bed file to genepred format.', flush=True)
    args = get_args_parser()

    print('Step1: input the ribocode.', flush=True)
    record = input_the_ribocode(args)

    print('Step2: output the genepred file.', flush=True)
    output_genepred(record, args)

    print('Done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_odd_ratio.py


from .ribo import ArgsParser
from .ribo import Odd_Ratio


def main():
    ArgsParser.now_time()
    print('\nCalculate the relative codon pausing score.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_odd_ratio_args_parser()
    odd_ratio = Odd_Ratio.OddRatio(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    odd_ratio.read_rpf()

    print('\nStep3: Make two dimensional table for fisher exactly test.', flush=True)
    odd_ratio.make_two_dimensional_table()

    print('\nStep4: Calculate the odd ratio.', flush=True)
    odd_ratio.calc_odd_ratio()
    odd_ratio.calc_chi2_test2()

    print('\nStep5: Output the ribosome pausing odd ratio.', flush=True)
    odd_ratio.output_odd_ratio()
    odd_ratio.summarize_odd_ratio()
    
    print('\nStep6: Draw the ribosome pausing odd ratio.', flush=True)
    odd_ratio.draw_odd_ratio_line()
    odd_ratio.draw_odd_ratio_scatter()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : serp_overlap.py


import matplotlib
from matplotlib import pyplot as plt
from matplotlib_venn import venn2
import pandas as pd
from interval import Interval

from .ribo import ArgsParser


def import_peak_region(mock, flag):
    mock_peak = pd.read_csv(mock, sep='\t')
    sig_mock_peak = mock_peak[mock_peak['BHFDR'].replace('-', 'nan').astype(float) < 0.05].copy()

    flag_peak = pd.read_csv(flag, sep='\t')
    sig_flag_peak = flag_peak[flag_peak['BHFDR'].replace('-', 'nan').astype(float) < 0.05].copy()

    return sig_mock_peak, sig_flag_peak


def get_overlap(left_peak, right_peak, marker):
    for idx_mock, rows_mock in left_peak.iterrows():
        if rows_mock.overlap != 'NaN':
            continue
        # get the same genes in another peak table
        tmp_flag = right_peak[right_peak.transcripts == rows_mock.transcripts]
        if tmp_flag.empty:
            if left_peak.loc[idx_mock, 'overlap'] == 'NaN':
                left_peak.loc[idx_mock, 'overlap'] = marker
        else:
            # make the interval for test
            mock_start = int(rows_mock.peak_start)
            mock_end = int(rows_mock.peak_end)
            mock_interval = Interval(mock_start, mock_end, lower_closed=True, upper_closed=True)
            for idx_flag, rows_flag in tmp_flag.iterrows():
                # make the interval for test
                flag_start = int(rows_flag.peak_start)
                flag_end = int(rows_flag.peak_end)
                flag_interval = Interval(flag_start, flag_end, lower_closed=True, upper_closed=True)
                # annotate the overlap condition
                if flag_interval.overlaps(mock_interval):
                    if left_peak.loc[idx_mock, 'overlap'] == 'NaN':
                        left_peak.loc[idx_mock, 'overlap'] = 'overlap'
                else:
                    if left_peak.loc[idx_mock, 'overlap'] == 'NaN':
                        left_peak.loc[idx_mock, 'overlap'] = marker
    return left_peak


def merge_overlap(sig_mock_peak, sig_flag_peak, out_mock, out_flag):
    # add 'overlap' column in to the peak table
    sig_mock_peak.insert(len(sig_mock_peak.columns), 'overlap', 'NaN', allow_duplicates=False)
    sig_flag_peak.insert(len(sig_flag_peak.columns), 'overlap', 'NaN', allow_duplicates=False)
    # get the overlap region and annotation
    sig_mock_anno = get_overlap(sig_mock_peak, sig_flag_peak, 'mock-IP')
    sig_flag_anno = get_overlap(sig_flag_peak, sig_mock_peak, 'flag-IP')
    # output
    sig_mock_anno.to_csv(out_mock + '.overlap.txt', sep='\t', header=True, index=False)
    sig_flag_anno.to_csv(out_flag + '.overlap.txt', sep='\t', header=True, index=False)

    return sig_mock_anno, sig_flag_anno


def plot_venn2(sig_mock_anno, sig_flag_anno, out_mock, out_flag):
    sig_mock_gene = set(sig_mock_anno.transcripts)
    sig_flag_gene = set(sig_flag_anno.transcripts)

    sig_peak_num = pd.concat([sig_mock_anno.overlap, sig_flag_anno.overlap]).value_counts()
    # sometimes there will be multiple crossovers in the overlap region between two samples.
    # so use the average instead
    sig_peak_num.overlap = int(sig_peak_num.overlap / 2)

    pdf_file = out_mock + '_vs_' + out_flag + '_venn.pdf'
    png_file = out_mock + '_vs_' + out_flag + '_venn.png'

    matplotlib.use('AGG')
    fig = plt.figure(figsize=(8, 4), dpi=300)
    # draw the gene overlap figure
    ax1 = fig.add_subplot(1, 2, 1)
    venn2([sig_mock_gene, sig_flag_gene], set_labels=['mock-IP', 'flag-IP'])
    plt.title('gene overlap')
    # draw the peak region overlap figure
    ax2 = fig.add_subplot(1, 2, 2)
    venn2([sig_peak_num['mock-IP'], sig_peak_num['flag-IP'], sig_peak_num['overlap']], set_labels=['mock-IP', 'flag-IP'])
    plt.title('peak region overlap')
    # output the figure
    fig.tight_layout()
    # plt.show()
    fig.savefig(fname=pdf_file)
    fig.savefig(fname=png_file, dpi=300)


def main():
    ArgsParser.now_time()
    print('\nRetrieve the sequence of peak region.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.serp_overlap()

    print('\nStep2: Import the peak region.\n', flush=True)
    sig_mock_peak, sig_flag_peak = import_peak_region(args.mock, args.flag)

    print('\nStep3: Get the overlap region.\n', flush=True)
    if not args.out_mock:
        out_mock = args.mock[::-4]
    else:
        out_mock = args.out_mock
    if not args.out_flag:
        out_flag = args.flag[::-4]
    else:
        out_flag = args.out_flag
    sig_mock_anno, sig_flag_anno = merge_overlap(sig_mock_peak, sig_flag_peak, out_mock, out_flag)

    print('\nStep4: Draw the venn figure.\n', flush=True)
    plot_venn2(sig_mock_anno, sig_flag_anno, args.out_mock, args.out_flag)

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Quality.py


import os.path

import random
import sys
from collections import Counter
from collections import OrderedDict
from itertools import chain
from itertools import islice

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pysam
import seaborn as sns

from multiprocessing import Pool

class Quality(object):

    def __init__(self, args):
        # opts for mrna file import
        self.longest = args.longest
        self.mrna_file = args.transcript
        self.mrna_dict = OrderedDict()
        self.length_dict = {}

        # opts for bam file import
        self.thread = args.thread
        self.pysam_input = None
        self.pysam_output = None
        self.sample_file = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]

        self.bam_seq_dict = {}
        self.sample_dict = None
        self.tag = args.tag
        self.reverse = args.reverse
        self.align = args.align

        # opts for bam length summary
        self.peak_length = 0
        self.peak_reads = 0
        self.mono = 1
        self.profile = None

        # opts for rpf saturation
        self.saturation_flag = args.saturation
        self.saturation = []
        self.x_ticks = np.array([i * 10 for i in range(1, 10)])

        # opts for file output
        self.output_prefix = args.output

    def read_transcript(self):
        '''
        @Message  : read the transcript file and store the information into a dict.
        @Input    : self.mrna_file --> the path of transcript file
        @Return   : self.mrna_dict --> the dict of transcript information
        @Flow     : step1: open the transcript file
                    step2: read the transcript file and store the information into a dict
        '''
        
        trans_file = self.mrna_file
        with open(trans_file, 'r') as trans_file_in:
            for line in islice(trans_file_in, 1, None):
                record = line.strip().split('\t')
                self.mrna_dict[record[2]] = [0] * 9
    
    def sort_index_bam(self):
        # check the data type, to support the SAM/BAM file format both.
        if self.sample_format.lower() == ".bam":
            print("import file: {bam}.\n".format(bam=self.sample_file), flush=True)
            self.sample_format = 'rb'
        elif self.sample_format.lower() == ".sam":
            print("import file: {bam}.\n".format(bam=self.sample_file), flush=True)
            self.sample_format = 'r'
        else:
            print(
                "Unknown file format, please input the correct bam or sam file.", flush=True)
            sys.exit()
        
        # check the bam index file
        if os.path.exists(self.sample_file + ".bai"):
            pass
        else:
            print("index file: {bai} doesn't exist, create the index file.\n".format(bai=self.sample_file + ".bai"), flush=True)
            pysam.sort("-o", self.output_prefix + ".temp.sorted.bam", self.sample_file, "-@", str(self.thread))
            pysam.index(self.output_prefix + ".temp.sorted.bam", "-@", str(self.thread))
            self.sample_file = self.output_prefix + ".temp.sorted.bam"

    @staticmethod
    def flt_star_results(args):
        '''
        @Message  : filter the bam file generated by STAR
        @Input    : bam_in_file --> the path of input bam file
                    bam_out_file --> the path of output bam file
                    gene_list --> the list of gene name
                    tag --> the number of unique reads
                    sample_format --> the format of bam file
        @Return   : length_dict --> the dict of length distribution
                    bam_seq_dict --> the dict of sequence profile
        @Flow     : step1: open the bam file
                    step2: read the bam file and store the information into a dict
                    step3: calculate the length distribution
                    step4: output the filtered bam file
        '''
        
        bam_in_file, bam_out_file, gene_list, tag, reverse_flag, sample_format = args

        tag_num = 0
        if tag == 0:
            tag_num = 99
        elif tag == 1:
            tag_num = 1
        
        bam_seq_dict = {}
        length_dict = {}
        
        with pysam.AlignmentFile(bam_in_file, sample_format) as bam_in:
            with pysam.AlignmentFile(bam_out_file, 'wb', template=bam_in) as bam_out:
                for gene in gene_list:
                    try:
                        gene_reads = bam_in.fetch(gene)
                    except ValueError:
                        continue

                    # filter the mapped reads or not
                    for reads in gene_reads:

                        # if reads.query_name == "A00202:1499:HWGTNDSX7:3:2518:22914:34194":
                        #     print(reads)
                        #     pass

                        if reads.is_unmapped:
                            continue

                        # filter the unique reads
                        if reads.has_tag('NH') and reads.get_tag('NH') <= tag_num:
                            read_length = reads.infer_read_length()
                            if reads.is_reverse:
                                try:
                                    length_dict[read_length][1] += 1
                                except KeyError:
                                    length_dict[read_length] = [0, 1]

                                if reverse_flag:
                                    try:
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)
                                    except KeyError:
                                        bam_seq_dict[reads.query_name] = set()
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)

                            else:
                                try:
                                    length_dict[read_length][0] += 1
                                except KeyError:
                                    length_dict[read_length] = [1, 0]
                                # mrna_dict[reads.reference_name].bam.append(reads)
                                try:
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)
                                except KeyError:
                                    bam_seq_dict[reads.query_name] = set()
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)
                bam_out.close()
            bam_in.close()

        return length_dict, bam_seq_dict
    
    @staticmethod
    def flt_hisat2_results(args):
        '''
        @Message  : filter the bam file generated by STAR
        @Input    : bam_in_file --> the path of input bam file
                    bam_out_file --> the path of output bam file
                    gene_list --> the list of gene name
                    tag --> the number of unique reads
                    sample_format --> the format of bam file
        @Return   : length_dict --> the dict of length distribution
                    bam_seq_dict --> the dict of sequence profile
        @Flow     : step1: open the bam file
                    step2: read the bam file and store the information into a dict
                    step3: calculate the length distribution
                    step4: output the filtered bam file
        '''

        bam_in_file, bam_out_file, gene_list, tag, reverse_flag, sample_format = args

        tag_num = 0
        if tag == 0:
            zs_flg = True
            tag_num = len(gene_list)
        elif tag == 1:
            zs_flg = False
            tag_num = 1

        bam_seq_dict = {}
        length_dict = {}

        with pysam.AlignmentFile(bam_in_file, sample_format) as bam_in:
            with pysam.AlignmentFile(bam_out_file, 'wb', template=bam_in) as bam_out:
                for gene in gene_list:
                    try:
                        gene_reads = bam_in.fetch(gene)
                    except ValueError:
                        continue

                    # filter the mapped reads or not
                    for reads in gene_reads:
                        if reads.is_unmapped:
                            continue
                        # filter the unique reads
                        if reads.has_tag('NH') and reads.get_tag('NH') <= tag_num and reads.get_tag('ZS') == zs_flg:
                            read_length = reads.infer_read_length()
                            if reads.is_reverse:
                                try:
                                    length_dict[read_length][1] += 1
                                except KeyError:
                                    length_dict[read_length] = [0, 1]

                                if reverse_flag:
                                    try:
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)
                                    except KeyError:
                                        bam_seq_dict[reads.query_name] = set()
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)

                            else:
                                try:
                                    length_dict[read_length][0] += 1
                                except KeyError:
                                    length_dict[read_length] = [1, 0]
                                # mrna_dict[reads.reference_name].bam.append(reads)
                                try:
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)
                                except KeyError:
                                    bam_seq_dict[reads.query_name] = set()
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)

        return length_dict, bam_seq_dict
    
    @staticmethod
    def flt_bowtie2_results(args):
        '''
        @Message  : filter the bam file generated by STAR
        @Input    : bam_in_file --> the path of input bam file
                    bam_out_file --> the path of output bam file
                    gene_list --> the list of gene name
                    tag --> the number of unique reads
                    sample_format --> the format of bam file
        @Return   : length_dict --> the dict of length distribution
                    bam_seq_dict --> the dict of sequence profile
        @Flow     : step1: open the bam file
                    step2: read the bam file and store the information into a dict
                    step3: calculate the length distribution
                    step4: output the filtered bam file
        '''

        bam_in_file, bam_out_file, gene_list, tag, reverse_flag, sample_format = args

        tag_num = 0
        if tag == 0:
            zs_flg = True
            tag_num = len(gene_list)
        elif tag == 1:
            zs_flg = False
            tag_num = 1

        bam_seq_dict = {}
        length_dict = {}

        with pysam.AlignmentFile(bam_in_file, sample_format) as bam_in:
            with pysam.AlignmentFile(bam_out_file, 'wb', template=bam_in) as bam_out:
                for gene in gene_list:
                    try:
                        gene_reads = bam_in.fetch(gene)
                    except ValueError:
                        continue
                    
                    # filter the mapped reads or not
                    for reads in gene_reads:
                        if reads.is_unmapped:
                            continue
                        # filter the unique reads
                        if reads.has_tag('AS') and reads.get_tag('AS') <= tag_num and reads.get_tag('XS') == zs_flg:
                            read_length = reads.infer_read_length()
                            if reads.is_reverse:
                                try:
                                    length_dict[read_length][1] += 1
                                except KeyError:
                                    length_dict[read_length] = [0, 1]

                                if reverse_flag:
                                    try:
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)
                                    except KeyError:
                                        bam_seq_dict[reads.query_name] = set()
                                        bam_seq_dict[reads.query_name].add(reads.reference_name)
                                        bam_out.write(reads)

                            else:
                                try:
                                    length_dict[read_length][0] += 1
                                except KeyError:
                                    length_dict[read_length] = [1, 0]
                                # mrna_dict[reads.reference_name].bam.append(reads)
                                try:
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)
                                except KeyError:
                                    bam_seq_dict[reads.query_name] = set()
                                    bam_seq_dict[reads.query_name].add(reads.reference_name)
                                    bam_out.write(reads)
        
        return length_dict, bam_seq_dict

    def fliter_mrna_reads(self):
        '''
        @Message  : retrieve the corrected rpf reads from bam file
        @Input    : self.sample_file --> the input bam file
                    self.sample_format --> the input bam file format
        @Return   : self.pysam_input --> the input bam file
                    self.pysam_output --> the output bam file
        @Flow     : step1 --> check the bam file format
                    step2 --> check the reads aligner algorithm of bam file
                    step3 --> import the bam file
                    step4 --> return the input and output bam file
        '''

        # flt by different algorithm
        pool = Pool(processes=self.thread)
        # splits = [gene_list[i:i + gene_list_len // self.thread] for i in range(0, gene_list_len, ceil(gene_list_len / self.thread))]

        splits = np.array_split(list(self.mrna_dict.keys()), self.thread)
        splits = [list(split) for split in splits]

        if self.align.upper() == 'HISAT2':
            args = [(self.sample_file, self.output_prefix + '_split_' + str(i) + '.bam', split, self.tag, self.reverse, self.sample_format) for i, split in enumerate(splits)]
            results = pool.map(self.flt_hisat2_results, args)

        elif self.align.upper() == 'STAR':
            args = [(self.sample_file, self.output_prefix + '_split_' + str(i) + '.bam', split, self.tag, self.reverse, self.sample_format) for i, split in enumerate(splits)]
            results = pool.map(self.flt_star_results, args)

        elif self.align.upper() == 'BOWTIE2' or self.align.upper() == 'BOWTIE':
            args = [(self.sample_file, self.output_prefix + '_split_' + str(i) + '.bam', split, self.tag, self.reverse, self.sample_format) for i, split in enumerate(splits)]
            results = pool.map(self.flt_bowtie2_results, args)
        else:
            print("Unknown align algorithm, please input the correct align algorithm.", flush=True)
            sys.exit()

        pool.close()
        pool.join()

        # merge length_dict with pandas
        # merge bam_seq_dict by dict keys and values
        first = True

        for length_dict, bam_seq_dict in results:
            if first:
                self.length_dict.update(length_dict)
                self.bam_seq_dict.update(bam_seq_dict)
                first = False
            else:
                for key, value in length_dict.items():
                    try:
                        self.length_dict[key][0] += value[0]
                        self.length_dict[key][1] += value[1]
                    except KeyError:
                        self.length_dict[key] = value
                        
                if self.saturation_flag:
                    self.bam_seq_dict = {**self.bam_seq_dict, **bam_seq_dict}
                    # for key, value in bam_seq_dict.items():
                    #     self.bam_seq_dict.setdefault(key, set()).update(value)
                else:
                    pass

        del results

    def merge_sort_index_bam(self):
        # merge the bam file
        bam_in_list = [self.output_prefix + "_split_" + str(i) + ".bam" for i in range(self.thread)]
        merge_parameters = ["-f", "-@", str(self.thread), self.output_prefix + ".bam"] + bam_in_list
        pysam.merge(*merge_parameters)

        # remove the temp bam file
        if os.path.exists(self.output_prefix + ".temp.sorted.bam"):
            os.remove(self.output_prefix + ".temp.sorted.bam")

        if os.path.exists(self.output_prefix + ".temp.sorted.bam.bai"):
            os.remove(self.output_prefix + ".temp.sorted.bam.bai")
        
        for i in range(self.thread):
            os.remove(self.output_prefix + "_split_" + str(i) + ".bam")

        # sort and index the bam file
        pysam.sort("-o", self.output_prefix + ".bam", self.output_prefix + ".bam", "-@", str(self.thread))
        pysam.index(self.output_prefix + ".bam", "-@", str(self.thread))

    def detect_seq_type(self):
        # auto-detect the type of sequence profile

        self.peak_length, self.peak_reads = max(self.length_dict.items(), key=lambda length: length[1])

        if 23 < self.peak_length < 35:
            print("{bam} is detected to be monosome-seq.\n".format(bam=self.sample_file), flush=True)
            self.profile = "monosome"
            self.mono = [19.5, 40.5]
        elif 53 < self.peak_length < 65:
            print("{bam} is detected to be disome-seq.\n".format(bam=self.sample_file), flush=True)
            self.profile = "disome"
            self.mono = [49.5, 70.5]
        elif 83 < self.peak_length < 95:
            print("{bam} is detected to be disome-seq.\n".format(bam=self.sample_file), flush=True)
            self.profile = "trisome"
            self.mono = [79.5, 100.5]
        elif 35 < self.peak_length < 53 or 65 < self.peak_length < 83 or 95 < self.peak_length:
            print("Warning! {bam} doesn't fit the empirical length distribution.!\n".format(bam=self.sample_file),
                  flush=True)
            print(''' 
            Monosome RPFs peak length is usually ~ 30 nt.
            Please check the files and run the detect_offset.py with specified peak_length!''',
                  flush=True)
            # sys.stdout.writelines('''
            # [25, 34] for the monosome, RPFs peak is usually ~ 30 nt.
            # [55, 65] for the disome, RPFs peak is usually ~ 60 nt.
            # [85, 95] for the trisome, RPFs peak is usually ~ 90 nt.
            # Please check the files and run the detect_offset.py with specified peak_length!''')

    def draw_the_length_distr(self, sorted_length):
        # output figure name
        out_pdf = self.output_prefix + "_length_distribution.pdf"
        out_png = self.output_prefix + "_length_distribution.png"

        matplotlib.use('AGG')

        # draw the figure
        length_df = pd.DataFrame(sorted_length).T
        fig = plt.figure(figsize=(6, 6), dpi=300)

        ax1 = fig.add_subplot(2, 1, 1)
        # ax1.bar(length_df.index, length_df[0], width=0.8, color="#FF9900")
        # sns.barplot(x=length_df.index, y=length_df[0], color="#f86934")
        sns.lineplot(x=length_df.index, y=length_df[0], color="#FF9900")
        ax1.set_xticks(length_df.index)
        ax1.set_xticklabels(length_df.index)
        ax1.set_xlim(self.mono)
        # ax1.set_ylim([0, round(length_df[0].max() + abs(length_df[0].max()) * 0.1)])
        ax1.set_ylabel('RPFs number')
        ax1.set_xlabel('RPFs length (nt)')
        ax1.set_title('plus strand')
        plt.xticks(rotation=90)

        # draw the minus strand RPFs length distribution
        ax2 = fig.add_subplot(2, 1, 2)
        # ax2.bar(length_df.index, length_df[1], width=0.8, color="#0099FF")
        # sns.barplot(x=length_df.index, y=length_df[0], color="#f86934")
        sns.lineplot(x=length_df.index, y=length_df[1], color="#0099FF")
        ax2.set_xticks(length_df.index)
        ax2.set_xticklabels(length_df.index)
        ax2.set_xlim(self.mono)
        # ax2.set_ylim([0, round(length_df[1].max() * 1.1)])
        ax2.set_ylabel('RPFs number')
        ax2.set_xlabel('RPFs length (nt)')
        ax2.set_title('Minus strand')

        plt.xticks(rotation=90)
        plt.suptitle('RPFs length distribution')
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)
        plt.close()

    def write_length_distr(self):
        # write the length distribution.
        sorted_length = dict(sorted(self.length_dict.items(), key=lambda length: length[0], reverse=False))

        with open(self.output_prefix + "_length_distribution.txt", 'w') as length_out:
            length_out.writelines('\t'.join(["Length", "Plus", "Minus"]) + '\n')
            for reads_length, reads_num in sorted_length.items():
                length_out.writelines('\t'.join([str(reads_length), str(reads_num[0]), str(reads_num[1])]) + '\n')

        # draw the RPFs length distribution
        self.draw_the_length_distr(sorted_length)

    @staticmethod
    def sample_keys(step, keys):
        keys_list = list(keys)
        return random.sample(keys, step)
    
    def rpf_saturation(self):
        '''
        @Message  : calculate the RPFs saturation.
        @Input    : self.bam_seq_dict --> dict, {read_name: [gene1, gene2, ...]}
        @Return   : self.mrna_dict --> dict, {gene1: [site1, site2, ...]}
        @Flow     : step1 --> create the empty dict for resample steps and mrna count.
                    step2 --> create the function to resample the reads.
                    step3 --> runt the resample reads function.
                    step4 --> create the function to fill the empty mrna count dict.
                    step5 --> fill the empty mrna count dict.
        '''
        
        mapped_reads = len(self.bam_seq_dict)
        steps = [int(i * 0.05 * mapped_reads) for i in range(1, 10)]
        site = 0

        for step in steps:
            
            temp = random.sample(list(self.bam_seq_dict.keys()), step)
            temp_gene = list(chain.from_iterable(self.bam_seq_dict[i] for i in temp))

            temp_gene_dict = dict(Counter(temp_gene))

            gene_num = len(temp_gene_dict)
            self.saturation.append(gene_num)

            for gene, num in temp_gene_dict.items():
                self.mrna_dict[gene][site] = num

            site += 1

    @staticmethod
    def process_step(step_data):
        step, bam_seq_dict = step_data
        temp = random.sample(list(bam_seq_dict.keys()), step)
        temp_gene = []

        for i in temp:
            temp_gene.extend(bam_seq_dict[i])

        temp_gene_dict = dict(Counter(temp_gene))
        gene_num = len(temp_gene_dict)

        result = (gene_num, temp_gene_dict)
        return result

    def rpf_saturation_thread(self):
        from multiprocessing import Pool

        mapped_reads = len(self.bam_seq_dict)
        steps = [int(i * 0.05 * mapped_reads) for i in range(1, 10)]
        site = 0

        # Create a process pool
        pool = Pool(processes=self.thread)

        # Run the process_step function in parallel for each step
        step_data = [(step, self.bam_seq_dict) for step in steps]
        results = pool.map(self.process_step, step_data)

        for gene_num, temp_gene_dict in results:
            self.saturation.append(gene_num)

            for gene, num in temp_gene_dict.items():
                self.mrna_dict[gene][site] = num

            site += 1
            
    def draw_gene_saturation(self):
        out_pdf = self.output_prefix + "_gene_saturation.pdf"
        out_png = self.output_prefix + "_gene_saturation.png"
        out_gene = self.output_prefix + "_gene_saturation.txt"

        total_gene_num = len(self.mrna_dict)

        # convert the list to dataframe ,and set the title with 'gene'
        gene_df = pd.DataFrame(self.saturation + [total_gene_num], columns=['Count'])
        gene_df['Part'] = self.x_ticks.tolist() + [0]
        gene_df = gene_df[['Part', 'Count']]
        gene_df.to_csv(out_gene, sep='\t', index=False)

        matplotlib.use('AGG')

        fig = plt.figure(figsize=(8, 4), dpi=300)
        ax1 = fig.add_subplot(1, 2, 1)
        ax1.bar(self.x_ticks, self.saturation, width=3, color="#FF9900")
        ax1.plot(self.x_ticks, self.saturation, linewidth=0.8)
        plt.xticks(rotation=90)
        ax1.set_xlabel('reads proportion (%)')
        ax1.set_ylabel('gene number')
        plt.title('covered gene saturation')

        ax2 = fig.add_subplot(1, 2, 2)
        ax2.bar(self.x_ticks, [total_gene_num - i for i in self.saturation], width=3, color="#FF9900")
        ax2.plot(self.x_ticks, [total_gene_num - i for i in self.saturation], linewidth=0.8)
        plt.xticks(rotation=90)
        ax2.set_xlabel('reads proportion (%)')
        ax2.set_ylabel('gene number')
        plt.title('uncovered gene saturation')

        plt.tight_layout()
        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)
        plt.close()

    def draw_rpf_saturation(self):
        # output figure name
        out_pdf = self.output_prefix + "_reads_saturation.pdf"
        out_png = self.output_prefix + "_reads_saturation.png"
        out_rpf = self.output_prefix + "_reads_saturation.txt"
        

        mrna_df = pd.DataFrame(self.mrna_dict).T
        mrna_df.loc[:, 'mean'] = mrna_df.mean(axis=1)
        mrna_df = mrna_df.sort_values(['mean'], ascending=True)
        mrna_df.to_csv(out_rpf, sep='\t')

        quantile_list = mrna_df['mean'].quantile([0.25, 0.5, 0.75])
        mrna_0_25 = mrna_df[mrna_df['mean'] < quantile_list.iloc[0]]
        mrna_25_50 = mrna_df[(quantile_list.iloc[0] <= mrna_df['mean']) & (mrna_df['mean'] < quantile_list.iloc[1])]
        mrna_50_75 = mrna_df[(quantile_list.iloc[1] <= mrna_df['mean']) & (mrna_df['mean'] < quantile_list.iloc[2])]
        mrna_75_100 = mrna_df[quantile_list.iloc[2] <= mrna_df['mean']]

        # draw the figure
        matplotlib.use('AGG')
        colors = ['#ffc7ce', '#c6efce', '#ffeb9c', '#9cc3e5']
        flierprops = dict(marker='o', markersize=2, markerfacecolor='#c95859', markeredgecolor='none')
        fig = plt.figure(figsize=(17, 4), dpi=300)

        fig.add_subplot(1, 4, 1)
        ax1 = sns.boxplot(data=mrna_0_25.iloc[:, 0:9], color='#30adfe', flierprops=flierprops)
        ax1.set(xlabel='reads proportion (%)', ylabel='Reads count')
        ax1.set_yscale('log')
        ax1.set_xticklabels(self.x_ticks, rotation=90)

        fig.add_subplot(1, 4, 2)
        ax2 = sns.boxplot(data=mrna_25_50.iloc[:, 0:9], color='#30adfe', flierprops=flierprops)
        ax2.set(xlabel='reads proportion (%)', ylabel='Reads count')
        ax2.set_yscale('log')
        ax2.set_xticklabels(self.x_ticks, rotation=90)

        fig.add_subplot(1, 4, 3)
        ax3 = sns.boxplot(data=mrna_50_75.iloc[:, 0:9], color='#30adfe', flierprops=flierprops)
        ax3.set(xlabel='reads proportion (%)', ylabel='Reads count')
        ax3.set_yscale('log')
        ax3.set_xticklabels(self.x_ticks, rotation=90)

        fig.add_subplot(1, 4, 4)
        ax4 = sns.boxplot(data=mrna_75_100.iloc[:, 0:9], color='#30adfe', flierprops=flierprops)
        ax4.set(xlabel='reads proportion (%)', ylabel='Reads count')
        ax4.set_yscale('log')
        ax4.set_xticklabels(self.x_ticks, rotation=90)

        plt.suptitle('Reads saturation')
        plt.tight_layout()
        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)
        plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboparser
# @Script  : rpf_bam2bw.py


from .ribo import ArgsParser
from .ribo import Bam2Wig


def main():
    ArgsParser.now_time()
    print('\nConvert genome bam reads to bedgraph.\n', flush=True)
    print('Step1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_bam2bw_args_parser()

    bam_attr = Bam2Wig.Bam2Wig(args)

    print('\nStep2: Import the P-site offset.', flush=True)
    bam_attr.read_offset()

    print('\nStep3: Import the alignment bam file.', flush=True)
    bam_attr.set_tag_num()
    bam_attr.import_bam()

    print('\nStep4: convert bam reads to density.', flush=True)
    bam_attr.convert_dict_to_dataframe()

    print('\nStep5: Normalize the reads density.', flush=True)
    bam_attr.norm_rpm()

    print('\nStep6: Output reads density.', flush=True)
    bam_attr.output_bed()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2020/8/2 23:17
# @Project : py-scripts
# @Script  : metaBg.py
# @Version : python 3.8.5
# @product : PyCharm
# @Author  : Rensc
# @E-mail  : rensc0718@163.com


import argparse
import gc
import time
from argparse import RawTextHelpFormatter


# get arguments and define the scripts usage
def parse_args():
    parser = argparse.ArgumentParser(
        description="This script is used to convert the bedgraph to plot data.",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s 1.0"
    )
    parser.add_argument(
        "-i", metavar="input", required=True, type=str, help="input file"
    )
    parser.add_argument(
        "-o", metavar="output", required=True, type=str, help="output file"
    )
    # parser.add_argument(
    #     "-t", metavar="type", required=False, type=str, default='t',
    #     help="RPKM or TPM. (default: %(default)s)"
    # )

    args = parser.parse_args()
    return args


# convert the bedgraph to each nucleotide
def convert_bg(in_bg_file, out_bg_file):
    # pos_minus_list = []
    pos_plus_list = []
    read_minus_list = []
    read_plus_list = []
    with open(in_bg_file, 'r') as in_bg:
        for line in in_bg:
            rec = line.strip().split('\t')
            if rec[-1][0] == '-':
                pos_minus = [rec[0]+"\t"+str(i) for i in range(int(rec[1])+1, int(rec[2])+1)]
                read_minus = [rec[3] for i in range(int(rec[1]), int(rec[2]))]
                # pos_minus_list.extend(pos_minus)
                read_minus_list.extend(read_minus)
            else:
                pos_plus = [rec[0]+"\t"+str(i) for i in range(int(rec[1])+1, int(rec[2])+1)]
                read_plus = [rec[3] for i in range(int(rec[1]), int(rec[2]))]
                pos_plus_list.extend(pos_plus)
                read_plus_list.extend(read_plus)

    with open(out_bg_file, 'w') as out_bg:
        for line in range(len(pos_plus_list)):
            out_bg.writelines('\t'.join([pos_plus_list[line], read_plus_list[line]]) + '\n')
            # out_bg.writelines('\t'.join([pos_plus_list[line], read_plus_list[line], read_minus_list[line]]) + '\n')


# gc collect
def clear():
    print("Clear Memory.")
    gc.collect()
    print("All done.")


# get time
def now_time():
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))


# main
def main():
    now_time()
    args = parse_args()
    convert_bg(args.i, args.o)
    clear()
    now_time()


if __name__ == '__main__':
    main()
